import numpy as np
import pytest

from muGrid import Communicator, GlobalFieldCollection
from muGrid.Solvers import conjugate_gradients


@pytest.mark.skipif(Communicator().size > 1, reason="Test only works in serial")
@pytest.mark.parametrize(
    "A,b,x0",
    [
        ([[1, 0, 0], [0, 1, 0], [0, 0, 1]], [1, 2, 3], [0, 0, 0]),
        ([[1, 1, 0], [1, 2, 0], [0, 0, 1]], [1, 2, 3], [0, 0, 0]),
        ([[4, 1], [1, 3]], [1, 2], [2, 1]),
    ],
)
def test_conjugate_gradients(A, b, x0):
    comm = Communicator()

    fc = GlobalFieldCollection((len(b),))

    A = np.array(A)

    def hessp(x, Ax):
        Ax.p[...] = A @ x.p
        return Ax

    solution = fc.real_field("solution")
    solution.p[...] = np.array(x0)
    rhs = fc.real_field("rhs")
    rhs.p[...] = np.array(b)

    conjugate_gradients(
        comm, fc, hessp, rhs, solution, tol=1e-6, maxiter=10
    )

    np.testing.assert_allclose(solution.p, np.linalg.solve(A, b), atol=1e-6)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@file    python_laminate_homogenization_tests.py

@author  Lars Pastewka <lars.pastewka@imtek.uni-freiburg.de>

@date    02 Jan 2026

@brief   Tests for isotropic stiffness kernel using laminate homogenization.

This test verifies that the homogenized elastic properties of laminate
microstructures exactly match the analytical Voigt and Reuss bounds.

For a two-phase laminate:
- Voigt bound (uniform strain): applies when loading perpendicular to layers
- Reuss bound (uniform stress): applies when loading parallel to layers

Copyright 2026 Lars Pastewka

uGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

uGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with uGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

import numpy as np
import pytest
from NuMPI.Testing.Subdivision import suggest_subdivisions

import muGrid
from muGrid import Timer
from muGrid.Solvers import conjugate_gradients

# =============================================================================
# Helper functions
# =============================================================================


def voigt_index_2d(i, j):
    """Convert 2D tensor indices to Voigt notation index."""
    if i == j:
        return i
    return 2


def voigt_index_3d(i, j):
    """Convert 3D tensor indices to Voigt notation index.

    Voigt ordering: xx=0, yy=1, zz=2, yz=3, xz=4, xy=5
    """
    if i == j:
        return i
    pair = tuple(sorted([i, j]))
    mapping = {(1, 2): 3, (0, 2): 4, (0, 1): 5}
    return mapping[pair]


def voigt_index(dim, i, j):
    """Convert tensor indices to Voigt notation index for given dimension."""
    if dim == 2:
        return voigt_index_2d(i, j)
    else:
        return voigt_index_3d(i, j)


def isotropic_stiffness_2d(E, nu):
    """
    Create 2D plane strain isotropic stiffness tensor in Voigt notation.
    Returns C[3, 3] where [xx, yy, xy] ordering is used.
    """
    lam = E * nu / ((1 + nu) * (1 - 2 * nu))
    mu = E / (2 * (1 + nu))

    C = np.zeros((3, 3))
    C[0, 0] = lam + 2 * mu  # C_xxxx
    C[1, 1] = lam + 2 * mu  # C_yyyy
    C[2, 2] = mu  # C_xyxy
    C[0, 1] = C[1, 0] = lam  # C_xxyy
    return C


def isotropic_stiffness_3d(E, nu):
    """
    Create 3D isotropic stiffness tensor in Voigt notation.
    Returns C[6, 6] where [xx, yy, zz, yz, xz, xy] ordering is used.
    """
    lam = E * nu / ((1 + nu) * (1 - 2 * nu))
    mu = E / (2 * (1 + nu))

    C = np.zeros((6, 6))
    C[0, 0] = lam + 2 * mu
    C[1, 1] = lam + 2 * mu
    C[2, 2] = lam + 2 * mu
    C[3, 3] = mu
    C[4, 4] = mu
    C[5, 5] = mu
    C[0, 1] = C[1, 0] = lam
    C[0, 2] = C[2, 0] = lam
    C[1, 2] = C[2, 1] = lam
    return C


def compute_voigt_average(C1, C2, f1):
    """Compute Voigt (arithmetic) average of stiffness tensors."""
    f2 = 1.0 - f1
    return f1 * C1 + f2 * C2


def compute_reuss_average(C1, C2, f1):
    """Compute Reuss (harmonic) average of stiffness tensors."""
    f2 = 1.0 - f1
    S1 = np.linalg.inv(C1)
    S2 = np.linalg.inv(C2)
    S_eff = f1 * S1 + f2 * S2
    return np.linalg.inv(S_eff)


def laminate_stiffness_2d(E1, E2, nu, f1, layer_dir):
    """
    Compute exact laminate stiffness tensor for 2D plane strain.

    For a laminate with layers normal to `layer_dir`, the effective stiffness
    is NOT simply Voigt or Reuss average due to Poisson coupling.

    Parameters
    ----------
    E1, E2 : float
        Young's moduli of the two phases.
    nu : float
        Poisson's ratio (same for both phases).
    f1 : float
        Volume fraction of phase 1.
    layer_dir : int
        Direction normal to layers (0=x, 1=y).

    Returns
    -------
    C_eff : ndarray
        Effective stiffness tensor in Voigt notation (3x3).
    """
    f2 = 1.0 - f1

    # Lame parameters
    lam1 = E1 * nu / ((1 + nu) * (1 - 2 * nu))
    mu1 = E1 / (2 * (1 + nu))
    lam2 = E2 * nu / ((1 + nu) * (1 - 2 * nu))
    mu2 = E2 / (2 * (1 + nu))

    # Plane strain modulus M = lambda + 2*mu
    M1 = lam1 + 2 * mu1
    M2 = lam2 + 2 * mu2

    # Effective stiffness components
    # For layer normal in direction d:
    # - C_dd follows Reuss (stress continuity perpendicular to layers)
    # - C_tt follows modified Voigt with Poisson correction (parallel to layers)
    # - C_dt has off-diagonal coupling

    # Reuss for perpendicular direction (d)
    C_dd = 1.0 / (f1 / M1 + f2 / M2)

    # Modified formula for parallel direction (t)
    # C_tt = Voigt - f1*f2*(lambda1 - lambda2)^2 / (M1*f2 + M2*f1)
    M_voigt = f1 * M1 + f2 * M2
    correction = f1 * f2 * (lam1 - lam2) ** 2 / (M1 * f2 + M2 * f1)
    C_tt = M_voigt - correction

    # Off-diagonal coupling C_dt = C_td
    # Analytical: C_dt = weighted average of lambda with correction
    lam_voigt = f1 * lam1 + f2 * lam2
    lam_correction = f1 * f2 * (lam1 - lam2) * (M1 - M2) / (M1 * f2 + M2 * f1)
    C_dt = lam_voigt - lam_correction

    # Shear modulus: Reuss average (stress continuity for shear)
    mu_reuss = 1.0 / (f1 / mu1 + f2 / mu2)

    # Build stiffness tensor
    C_eff = np.zeros((3, 3))
    t_dir = 1 - layer_dir  # tangent direction

    C_eff[layer_dir, layer_dir] = C_dd
    C_eff[t_dir, t_dir] = C_tt
    C_eff[layer_dir, t_dir] = C_dt
    C_eff[t_dir, layer_dir] = C_dt
    C_eff[2, 2] = mu_reuss

    return C_eff


def laminate_stiffness_3d(E1, E2, nu, f1, layer_dir):
    """
    Compute exact laminate stiffness tensor for 3D.

    For a laminate with layers normal to `layer_dir`, the effective stiffness
    accounts for stress/strain continuity constraints.

    Parameters
    ----------
    E1, E2 : float
        Young's moduli of the two phases.
    nu : float
        Poisson's ratio (same for both phases).
    f1 : float
        Volume fraction of phase 1.
    layer_dir : int
        Direction normal to layers (0=x, 1=y, 2=z).

    Returns
    -------
    C_eff : ndarray
        Effective stiffness tensor in Voigt notation (6x6).
    """
    f2 = 1.0 - f1

    # Lame parameters
    lam1 = E1 * nu / ((1 + nu) * (1 - 2 * nu))
    mu1 = E1 / (2 * (1 + nu))
    lam2 = E2 * nu / ((1 + nu) * (1 - 2 * nu))
    mu2 = E2 / (2 * (1 + nu))

    # Plane strain modulus M = lambda + 2*mu
    M1 = lam1 + 2 * mu1
    M2 = lam2 + 2 * mu2

    # For layer normal in direction d and tangent directions t1, t2:
    # - C_dd: Reuss (stress continuity perpendicular to layers)
    # - C_t1t1, C_t2t2: Modified with Poisson correction
    # - Coupling terms adjusted

    # Reuss for perpendicular direction
    C_dd = 1.0 / (f1 / M1 + f2 / M2)

    # Modified for parallel directions
    M_voigt = f1 * M1 + f2 * M2
    correction = f1 * f2 * (lam1 - lam2) ** 2 / (M1 * f2 + M2 * f1)
    C_tt = M_voigt - correction

    # Coupling: lambda between d and t
    lam_voigt = f1 * lam1 + f2 * lam2
    lam_correction = f1 * f2 * (lam1 - lam2) * (M1 - M2) / (M1 * f2 + M2 * f1)
    C_dt = lam_voigt - lam_correction

    # Coupling: lambda between t1 and t2 (both parallel to layers)
    # This also has Poisson correction due to equilibrium constraint in layer normal
    # The correction is the same as for C_tt
    C_t1t2 = lam_voigt - correction

    # Shear moduli
    # Shear in plane containing d: Reuss
    mu_reuss = 1.0 / (f1 / mu1 + f2 / mu2)
    # Shear in plane parallel to layers: Voigt
    mu_voigt = f1 * mu1 + f2 * mu2

    # Build stiffness tensor in Voigt notation
    # Ordering: xx=0, yy=1, zz=2, yz=3, xz=4, xy=5
    C_eff = np.zeros((6, 6))

    # Get tangent directions
    tangent_dirs = [i for i in range(3) if i != layer_dir]
    t1, t2 = tangent_dirs

    # Normal components
    C_eff[layer_dir, layer_dir] = C_dd
    C_eff[t1, t1] = C_tt
    C_eff[t2, t2] = C_tt

    # Coupling between d and tangent
    C_eff[layer_dir, t1] = C_dt
    C_eff[t1, layer_dir] = C_dt
    C_eff[layer_dir, t2] = C_dt
    C_eff[t2, layer_dir] = C_dt

    # Coupling between tangent directions
    C_eff[t1, t2] = C_t1t2
    C_eff[t2, t1] = C_t1t2

    # Shear components (Voigt indices 3, 4, 5 correspond to yz, xz, xy)
    # Shear in plane containing layer_dir uses Reuss
    # Shear in plane parallel to layers uses Voigt
    for shear_idx in [3, 4, 5]:
        # Determine which plane this shear is in
        # 3=yz (plane perpendicular to x), 4=xz (perp to y), 5=xy (perp to z)
        perp_to = {3: 0, 4: 1, 5: 2}[shear_idx]
        if perp_to == layer_dir:
            # Shear parallel to layers
            C_eff[shear_idx, shear_idx] = mu_voigt
        else:
            # Shear involving layer normal direction
            C_eff[shear_idx, shear_idx] = mu_reuss

    return C_eff


# =============================================================================
# Homogenization solver
# =============================================================================


class LaminateHomogenization:
    """
    Compute homogenized stiffness of a laminate microstructure.

    The laminate consists of alternating layers of two isotropic materials.
    """

    def __init__(
        self,
        nb_grid_pts,
        E1,
        E2,
        nu,
        layer_direction,
        volume_fraction_1=0.5,
        tol=1e-6,
        maxiter=1000,
    ):
        """
        Initialize laminate homogenization.

        Parameters
        ----------
        nb_grid_pts : tuple
            Number of grid points in each direction.
        E1 : float
            Young's modulus of phase 1.
        E2 : float
            Young's modulus of phase 2.
        nu : float
            Poisson's ratio (same for both phases).
        layer_direction : int
            Direction normal to layers (0=x, 1=y, 2=z).
        volume_fraction_1 : float
            Volume fraction of phase 1 (default 0.5).
        tol : float
            CG tolerance.
        maxiter : int
            Maximum CG iterations.
        """
        self.nb_grid_pts = nb_grid_pts
        self.dim = len(nb_grid_pts)
        self.E1 = E1
        self.E2 = E2
        self.nu = nu
        self.layer_direction = layer_direction
        self.volume_fraction_1 = volume_fraction_1
        self.tol = tol
        self.maxiter = maxiter

        # Communicator
        try:
            from mpi4py import MPI

            self.comm = muGrid.Communicator(MPI.COMM_WORLD)
        except ImportError:
            self.comm = muGrid.Communicator()

        # Compute Lame parameters
        self.lam1 = E1 * nu / ((1 + nu) * (1 - 2 * nu))
        self.mu1 = E1 / (2 * (1 + nu))
        self.lam2 = E2 * nu / ((1 + nu) * (1 - 2 * nu))
        self.mu2 = E2 / (2 * (1 + nu))

        # Domain size (unit cube)
        self.domain_size = tuple(1.0 for _ in range(self.dim))
        self.grid_spacing = [L / n for L, n in zip(self.domain_size, nb_grid_pts)]

        # Create gradient operator
        self.gradient_op = muGrid.FEMGradientOperator(self.dim, list(self.grid_spacing))
        self.nb_quad = self.gradient_op.nb_quad_pts
        self.quad_weights = self.gradient_op.quadrature_weights

        # Compute MPI subdivisions
        nb_subdivisions = suggest_subdivisions(self.dim, self.comm.size)

        # Create decomposition with ghost cells
        self.decomposition = muGrid.CartesianDecomposition(
            self.comm,
            nb_grid_pts,
            nb_subdivisions=nb_subdivisions,
            nb_ghosts_left=(1,) * self.dim,
            nb_ghosts_right=(1,) * self.dim,
            nb_sub_pts={"quad": self.nb_quad},
        )

        # Create element decomposition for material fields
        self.element_decomposition = muGrid.CartesianDecomposition(
            self.comm,
            nb_grid_pts,
            nb_subdivisions=nb_subdivisions,
            nb_ghosts_left=(1,) * self.dim,
            nb_ghosts_right=(1,) * self.dim,
        )

        # Store local grid shape for array allocations
        self.local_grid_shape = self.decomposition.nb_subdomain_grid_pts

        # Create stiffness operator
        if self.dim == 2:
            self.stiffness_op = muGrid.IsotropicStiffnessOperator2D(
                list(self.grid_spacing)
            )
        else:
            self.stiffness_op = muGrid.IsotropicStiffnessOperator3D(
                list(self.grid_spacing)
            )

        # Create fields
        self._create_fields()
        self._setup_material()

    def _create_fields(self):
        """Create required fields."""
        # Displacement and force fields
        self.u_field = self.decomposition.real_field("displacement", (self.dim,))
        self.rhs_field = self.decomposition.real_field("rhs", (self.dim,))
        self.force_field = self.decomposition.real_field("force", (self.dim,))

        # Material fields
        self.lambda_field = self.element_decomposition.real_field("lambda")
        self.mu_field = self.element_decomposition.real_field("mu")

        # Gradient and stress fields for post-processing
        self.gradient_field = self.decomposition.real_field(
            "gradient", (self.dim, self.dim), "quad"
        )
        self.stress_field = self.decomposition.real_field(
            "stress", (self.dim, self.dim), "quad"
        )

        # Stiffness tensor field for generic operator (use local shape!)
        self.nb_voigt = 3 if self.dim == 2 else 6
        self.C_field = np.zeros(
            (self.nb_voigt, self.nb_voigt, self.nb_quad) + self.local_grid_shape
        )

    def _setup_material(self):
        """Set up laminate material distribution."""
        # Use local coordinates from decomposition (handles MPI domain decomposition)
        local_coords = self.decomposition.coords

        # Determine phase based on layer direction
        # Phase 1 where coordinate < volume_fraction_1
        layer_coord = local_coords[self.layer_direction]
        phase = (layer_coord < self.volume_fraction_1).astype(float)

        # Set material properties
        self.lambda_field.p[...] = self.lam1 * phase + self.lam2 * (1 - phase)
        self.mu_field.p[...] = self.mu1 * phase + self.mu2 * (1 - phase)

        # Fill ghost cells
        self.element_decomposition.communicate_ghosts(self.lambda_field)
        self.element_decomposition.communicate_ghosts(self.mu_field)

        # Also set up stiffness tensor for post-processing
        C1 = (
            isotropic_stiffness_2d(self.E1, self.nu)
            if self.dim == 2
            else isotropic_stiffness_3d(self.E1, self.nu)
        )
        C2 = (
            isotropic_stiffness_2d(self.E2, self.nu)
            if self.dim == 2
            else isotropic_stiffness_3d(self.E2, self.nu)
        )

        for q in range(self.nb_quad):
            for i in range(self.nb_voigt):
                for j in range(self.nb_voigt):
                    self.C_field[i, j, q, ...] = C1[i, j] * phase + C2[i, j] * (
                        1 - phase
                    )

    def _apply_stiffness(self, u_in, f_out):
        """Apply stiffness operator K = B^T C B."""
        self.decomposition.communicate_ghosts(u_in)
        self.stiffness_op.apply(u_in, self.lambda_field, self.mu_field, f_out)

    def _strain_to_voigt(self, strain):
        """Convert strain tensor to Voigt notation."""
        if self.dim == 2:
            eps_voigt = np.zeros((3, self.nb_quad) + self.local_grid_shape)
            eps_voigt[0, ...] = strain[0, 0, ...]
            eps_voigt[1, ...] = strain[1, 1, ...]
            eps_voigt[2, ...] = 2 * strain[0, 1, ...]
        else:
            eps_voigt = np.zeros((6, self.nb_quad) + self.local_grid_shape)
            eps_voigt[0, ...] = strain[0, 0, ...]
            eps_voigt[1, ...] = strain[1, 1, ...]
            eps_voigt[2, ...] = strain[2, 2, ...]
            eps_voigt[3, ...] = 2 * strain[1, 2, ...]
            eps_voigt[4, ...] = 2 * strain[0, 2, ...]
            eps_voigt[5, ...] = 2 * strain[0, 1, ...]
        return eps_voigt

    def _voigt_to_stress(self, sig_voigt, stress):
        """Convert stress from Voigt notation to tensor."""
        if self.dim == 2:
            stress[0, 0, ...] = sig_voigt[0, ...]
            stress[1, 1, ...] = sig_voigt[1, ...]
            stress[0, 1, ...] = sig_voigt[2, ...]
            stress[1, 0, ...] = sig_voigt[2, ...]
        else:
            stress[0, 0, ...] = sig_voigt[0, ...]
            stress[1, 1, ...] = sig_voigt[1, ...]
            stress[2, 2, ...] = sig_voigt[2, ...]
            stress[1, 2, ...] = sig_voigt[3, ...]
            stress[2, 1, ...] = sig_voigt[3, ...]
            stress[0, 2, ...] = sig_voigt[4, ...]
            stress[2, 0, ...] = sig_voigt[4, ...]
            stress[0, 1, ...] = sig_voigt[5, ...]
            stress[1, 0, ...] = sig_voigt[5, ...]

    def _compute_stress(self, strain, stress):
        """Compute stress from strain using Voigt notation."""
        eps_voigt = self._strain_to_voigt(strain)
        sig_voigt = np.einsum("ijq...,jq...->iq...", self.C_field, eps_voigt)
        self._voigt_to_stress(sig_voigt, stress)

    def _compute_divergence(self, stress, f_vec):
        """Compute divergence of stress tensor."""
        self.stress_field.s[...] = stress
        self.decomposition.communicate_ghosts(self.stress_field)
        f_vec.pg[...] = 0.0
        self.gradient_op.transpose(self.stress_field, f_vec, list(self.quad_weights))

    def _compute_rhs(self, E_macro, rhs_out):
        """Compute RHS: f = -B^T C E_macro."""
        strain_shape = (self.dim, self.dim, self.nb_quad) + self.local_grid_shape
        eps_macro = np.zeros(strain_shape)
        for i in range(self.dim):
            for j in range(self.dim):
                eps_macro[i, j, ...] = E_macro[i, j]

        sig_macro = np.zeros_like(eps_macro)
        self._compute_stress(eps_macro, sig_macro)
        self._compute_divergence(sig_macro, rhs_out)
        rhs_out.s[...] *= -1.0

    def compute_homogenized_stiffness(self):
        """
        Compute the full homogenized stiffness tensor.

        Returns
        -------
        C_eff : ndarray
            Homogenized stiffness tensor in Voigt notation.
        """
        timer = Timer()
        C_eff = np.zeros((self.nb_voigt, self.nb_voigt))

        # Strain cases
        if self.dim == 2:
            strain_cases = [(0, 0), (1, 1), (0, 1)]
        else:
            strain_cases = [(0, 0), (1, 1), (2, 2), (1, 2), (0, 2), (0, 1)]

        for case_idx, (i, j) in enumerate(strain_cases):
            E_macro = np.zeros((self.dim, self.dim))
            if i == j:
                # Normal strain: tensor strain = 1.0
                E_macro[i, j] = 1.0
            else:
                # Shear strain: tensor strain = 0.5 so engineering shear = 1.0
                # In Voigt notation, C relates engineering shear (γ = 2ε) to stress
                E_macro[i, j] = 0.5
                E_macro[j, i] = 0.5

            voigt_col = voigt_index(self.dim, i, j)

            # Compute RHS
            self._compute_rhs(E_macro, self.rhs_field)

            # Initialize displacement
            self.u_field.s[...] = 0.0

            # Solve
            conjugate_gradients(
                self.comm,
                self.decomposition,
                self.rhs_field,
                self.u_field,
                hessp=self._apply_stiffness,
                tol=self.tol,
                maxiter=self.maxiter,
                timer=timer,
            )

            # Compute strain from solution
            self.decomposition.communicate_ghosts(self.u_field)
            self.gradient_op.apply(self.u_field, self.gradient_field)
            grad = self.gradient_field.s

            # Symmetric strain (use local shape!)
            strain_shape = (self.dim, self.dim, self.nb_quad) + self.local_grid_shape
            strain = np.zeros(strain_shape)
            for ii in range(self.dim):
                for jj in range(self.dim):
                    strain[ii, jj, ...] = 0.5 * (grad[ii, jj, ...] + grad[jj, ii, ...])

            # Add macroscopic strain
            for ii in range(self.dim):
                for jj in range(self.dim):
                    strain[ii, jj, ...] += E_macro[ii, jj]

            # Compute stress
            stress = np.zeros_like(strain)
            self._compute_stress(strain, stress)

            # Average stress
            sig_avg = np.zeros((self.dim, self.dim))
            for k in range(self.dim):
                for L in range(self.dim):
                    local_sum = 0.0
                    for q in range(self.nb_quad):
                        local_sum += self.quad_weights[q] * np.sum(stress[k, L, q, ...])
                    sig_avg[k, L] = self.comm.sum(local_sum)

            total_volume = np.prod(self.domain_size)
            sig_avg /= total_volume

            # Store in C_eff
            for k in range(self.dim):
                for L in range(self.dim):
                    voigt_row = voigt_index(self.dim, k, L)
                    if k <= L:
                        C_eff[voigt_row, voigt_col] = sig_avg[k, L]

        return C_eff


# =============================================================================
# Tests
# =============================================================================


class TestLaminateExactStiffness:
    """
    Test that computed laminate stiffness matches exact analytical formulas.

    For a laminate, the effective stiffness is NOT simply Voigt or Reuss average
    due to Poisson coupling effects. The exact analytical formulas account for
    stress/strain continuity constraints at layer interfaces.
    """

    @pytest.mark.parametrize("layer_dir", [0, 1])
    def test_exact_stiffness_2d(self, layer_dir):
        """Test full stiffness tensor against exact formula in 2D."""
        E1, E2 = 1.0, 10.0
        nu = 0.3
        f1 = 0.5
        nb_grid_pts = (16, 16)

        homog = LaminateHomogenization(
            nb_grid_pts, E1, E2, nu, layer_direction=layer_dir, volume_fraction_1=f1
        )
        C_eff = homog.compute_homogenized_stiffness()

        # Compute exact analytical laminate stiffness
        C_analytical = laminate_stiffness_2d(E1, E2, nu, f1, layer_dir)

        np.testing.assert_allclose(
            C_eff,
            C_analytical,
            rtol=1e-4,
            atol=1e-8,  # Handle tiny numerical errors near zero
            err_msg=f"Exact laminate stiffness failed for 2D, layer_dir={layer_dir}",
        )

    @pytest.mark.parametrize("layer_dir", [0, 1, 2])
    def test_exact_stiffness_3d(self, layer_dir):
        """Test full stiffness tensor against exact formula in 3D."""
        E1, E2 = 1.0, 10.0
        nu = 0.3
        f1 = 0.5
        nb_grid_pts = (8, 8, 8)

        homog = LaminateHomogenization(
            nb_grid_pts, E1, E2, nu, layer_direction=layer_dir, volume_fraction_1=f1
        )
        C_eff = homog.compute_homogenized_stiffness()

        # Compute exact analytical laminate stiffness
        C_analytical = laminate_stiffness_3d(E1, E2, nu, f1, layer_dir)

        np.testing.assert_allclose(
            C_eff,
            C_analytical,
            rtol=1e-4,
            atol=1e-8,  # Handle tiny numerical errors near zero
            err_msg=f"Exact laminate stiffness failed for 3D, layer_dir={layer_dir}",
        )


class TestLaminateReussBound:
    """
    Test that laminate with loading perpendicular to layers gives Reuss bound.

    For a laminate with layers normal to direction d, applying strain in
    direction d (perpendicular to layers) requires uniform stress at interfaces,
    giving Reuss bound.

    Physics:
    - Layers with normals in direction d stack perpendicular to d
    - Loading perpendicular to layers means stress must be continuous (equilibrium)
    - This gives Reuss (uniform stress / harmonic) average
    """

    @pytest.mark.parametrize("layer_dir", [0, 1])
    def test_reuss_bound_2d(self, layer_dir):
        """Test Reuss bound in 2D.

        For layers normal to layer_dir, loading in layer_dir (perpendicular to layers)
        should give Reuss bound.
        """
        E1, E2 = 1.0, 10.0
        nu = 0.3
        f1 = 0.5
        nb_grid_pts = (16, 16)

        homog = LaminateHomogenization(
            nb_grid_pts, E1, E2, nu, layer_direction=layer_dir, volume_fraction_1=f1
        )
        C_eff = homog.compute_homogenized_stiffness()

        C1 = isotropic_stiffness_2d(E1, nu)
        C2 = isotropic_stiffness_2d(E2, nu)
        C_reuss = compute_reuss_average(C1, C2, f1)

        # Loading perpendicular to layers should give Reuss bound
        np.testing.assert_allclose(
            C_eff[layer_dir, layer_dir],
            C_reuss[layer_dir, layer_dir],
            rtol=1e-4,
            err_msg=f"Reuss bound failed for 2D laminate, layer_dir={layer_dir}",
        )

    @pytest.mark.parametrize("layer_dir", [0, 1, 2])
    def test_reuss_bound_3d(self, layer_dir):
        """Test Reuss bound in 3D."""
        E1, E2 = 1.0, 10.0
        nu = 0.3
        f1 = 0.5
        nb_grid_pts = (8, 8, 8)

        homog = LaminateHomogenization(
            nb_grid_pts, E1, E2, nu, layer_direction=layer_dir, volume_fraction_1=f1
        )
        C_eff = homog.compute_homogenized_stiffness()

        C1 = isotropic_stiffness_3d(E1, nu)
        C2 = isotropic_stiffness_3d(E2, nu)
        C_reuss = compute_reuss_average(C1, C2, f1)

        np.testing.assert_allclose(
            C_eff[layer_dir, layer_dir],
            C_reuss[layer_dir, layer_dir],
            rtol=1e-4,
            err_msg=f"Reuss bound failed for 3D laminate, layer_dir={layer_dir}",
        )


class TestLaminateShearModulus:
    """
    Test shear modulus of laminate.

    For isotropic phases, shear modulus follows specific mixing rules
    depending on shear plane orientation relative to layers.

    For a laminate with layers normal to x:
    - Shear in xy or xz plane (involves x direction) → Reuss average
    - Shear in yz plane (parallel to layers) → Voigt average
    """

    def test_shear_involving_layer_normal_2d(self):
        """Test shear modulus for shear involving the layer normal direction in 2D.

        For layers normal to x (layer_dir=0), shear in xy plane involves the
        layer normal direction, so we expect Reuss (harmonic) average.
        """
        E1, E2 = 1.0, 10.0
        nu = 0.3
        f1 = 0.5
        nb_grid_pts = (16, 16)

        # Layers normal to x (layer_dir=0)
        homog = LaminateHomogenization(
            nb_grid_pts, E1, E2, nu, layer_direction=0, volume_fraction_1=f1
        )
        C_eff = homog.compute_homogenized_stiffness()

        # Shear moduli
        mu1 = E1 / (2 * (1 + nu))
        mu2 = E2 / (2 * (1 + nu))

        # For shear in xy plane with layers normal to x:
        # Shear stress continuity at interfaces → Reuss averaging
        mu_reuss = 1.0 / (f1 / mu1 + (1 - f1) / mu2)

        # C[2,2] is the shear modulus in 2D Voigt notation
        np.testing.assert_allclose(
            C_eff[2, 2],
            mu_reuss,
            rtol=1e-4,
            err_msg="Shear modulus (Reuss) failed for laminate with layer_dir=0",
        )


class TestLaminateSymmetry:
    """Test that homogenized stiffness tensor has correct symmetry."""

    def test_major_symmetry_2d(self):
        """Test major symmetry C_ijkl = C_klij in 2D."""
        E1, E2 = 1.0, 5.0
        nu = 0.3
        nb_grid_pts = (16, 16)

        homog = LaminateHomogenization(
            nb_grid_pts, E1, E2, nu, layer_direction=0, volume_fraction_1=0.5
        )
        C_eff = homog.compute_homogenized_stiffness()

        # Check symmetry with tolerance for numerical errors
        np.testing.assert_allclose(
            C_eff,
            C_eff.T,
            rtol=1e-6,
            atol=1e-8,
            err_msg="C_eff is not symmetric in 2D",
        )

    def test_major_symmetry_3d(self):
        """Test major symmetry C_ijkl = C_klij in 3D."""
        E1, E2 = 1.0, 5.0
        nu = 0.3
        nb_grid_pts = (8, 8, 8)

        homog = LaminateHomogenization(
            nb_grid_pts, E1, E2, nu, layer_direction=0, volume_fraction_1=0.5
        )
        C_eff = homog.compute_homogenized_stiffness()

        np.testing.assert_allclose(
            C_eff,
            C_eff.T,
            rtol=1e-6,
            atol=1e-8,
            err_msg="C_eff is not symmetric in 3D",
        )


class TestLaminatePositiveDefinite:
    """Test that homogenized stiffness tensor is positive definite."""

    def test_positive_definite_2d(self):
        """Test positive definiteness in 2D."""
        E1, E2 = 1.0, 10.0
        nu = 0.3
        nb_grid_pts = (16, 16)

        homog = LaminateHomogenization(
            nb_grid_pts, E1, E2, nu, layer_direction=0, volume_fraction_1=0.5
        )
        C_eff = homog.compute_homogenized_stiffness()

        eigenvalues = np.linalg.eigvalsh(C_eff)
        assert np.all(
            eigenvalues > 0
        ), f"C_eff not positive definite, eigenvalues: {eigenvalues}"

    def test_positive_definite_3d(self):
        """Test positive definiteness in 3D."""
        E1, E2 = 1.0, 10.0
        nu = 0.3
        nb_grid_pts = (8, 8, 8)

        homog = LaminateHomogenization(
            nb_grid_pts, E1, E2, nu, layer_direction=0, volume_fraction_1=0.5
        )
        C_eff = homog.compute_homogenized_stiffness()

        eigenvalues = np.linalg.eigvalsh(C_eff)
        assert np.all(
            eigenvalues > 0
        ), f"C_eff not positive definite, eigenvalues: {eigenvalues}"


class TestLaminateBoundsOrdering:
    """Test that effective properties lie between Voigt and Reuss bounds."""

    def test_bounds_ordering_2d(self):
        """Test bounds ordering in 2D."""
        E1, E2 = 1.0, 10.0
        nu = 0.3
        f1 = 0.5
        nb_grid_pts = (16, 16)

        homog = LaminateHomogenization(
            nb_grid_pts, E1, E2, nu, layer_direction=0, volume_fraction_1=f1
        )
        C_eff = homog.compute_homogenized_stiffness()

        C1 = isotropic_stiffness_2d(E1, nu)
        C2 = isotropic_stiffness_2d(E2, nu)
        C_voigt = compute_voigt_average(C1, C2, f1)
        C_reuss = compute_reuss_average(C1, C2, f1)

        # All diagonal components should be between Reuss and Voigt
        # Use small tolerance for numerical precision
        tol = 1e-8
        for i in range(3):
            assert C_reuss[i, i] - tol <= C_eff[i, i] <= C_voigt[i, i] + tol, (
                f"Bounds violated for C[{i},{i}]: "
                f"Reuss={C_reuss[i, i]}, Eff={C_eff[i, i]}, Voigt={C_voigt[i, i]}"
            )

    def test_bounds_ordering_3d(self):
        """Test bounds ordering in 3D."""
        E1, E2 = 1.0, 10.0
        nu = 0.3
        f1 = 0.5
        nb_grid_pts = (8, 8, 8)

        homog = LaminateHomogenization(
            nb_grid_pts, E1, E2, nu, layer_direction=0, volume_fraction_1=f1
        )
        C_eff = homog.compute_homogenized_stiffness()

        C1 = isotropic_stiffness_3d(E1, nu)
        C2 = isotropic_stiffness_3d(E2, nu)
        C_voigt = compute_voigt_average(C1, C2, f1)
        C_reuss = compute_reuss_average(C1, C2, f1)

        for i in range(6):
            assert C_reuss[i, i] - 1e-10 <= C_eff[i, i] <= C_voigt[i, i] + 1e-10, (
                f"Bounds violated for C[{i},{i}]: "
                f"Reuss={C_reuss[i, i]}, Eff={C_eff[i, i]}, Voigt={C_voigt[i, i]}"
            )


class TestHomogeneousMaterial:
    """Test that homogeneous material gives exact input properties."""

    def test_homogeneous_2d(self):
        """Test homogeneous material in 2D."""
        E = 10.0
        nu = 0.3
        nb_grid_pts = (16, 16)

        # Same E for both phases = homogeneous
        homog = LaminateHomogenization(
            nb_grid_pts, E, E, nu, layer_direction=0, volume_fraction_1=0.5
        )
        C_eff = homog.compute_homogenized_stiffness()

        C_analytical = isotropic_stiffness_2d(E, nu)

        np.testing.assert_allclose(
            C_eff,
            C_analytical,
            rtol=1e-6,
            err_msg="Homogeneous material does not match analytical",
        )

    def test_homogeneous_3d(self):
        """Test homogeneous material in 3D."""
        E = 10.0
        nu = 0.3
        nb_grid_pts = (8, 8, 8)

        homog = LaminateHomogenization(
            nb_grid_pts, E, E, nu, layer_direction=0, volume_fraction_1=0.5
        )
        C_eff = homog.compute_homogenized_stiffness()

        C_analytical = isotropic_stiffness_3d(E, nu)

        np.testing.assert_allclose(
            C_eff,
            C_analytical,
            rtol=1e-6,
            err_msg="Homogeneous material does not match analytical",
        )


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



import numpy as np
import muGrid

# Two dimensional grid
nb_grid_pts = (11, 12)
fc = muGrid.GlobalFieldCollection(nb_grid_pts)

# Get a scalar field
phase = fc.real_field("phase")
# The 1 is the number of sub points
assert phase.shape == [1, *nb_grid_pts]

# Get a single component field
single = fc.real_field("single", (1,))
# Now there is an explicit component dimension of 1
assert single.shape == [1, 1, *nb_grid_pts]



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@file    python_fem_gradient_gpu_tests.py

@author  Lars Pastewka <lars.pastewka@imtek.uni-freiburg.de>

@date    25 Dec 2024

@brief   Test GPU FEM gradient operator functionality (CUDA/ROCm)

Copyright (c) 2024 Lars Pastewka

muGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

muGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with muGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

import numpy as np
import pytest

# Import GPU testing utilities from conftest
from conftest import (
    HAS_CUPY,
    cp,
    create_device,
    get_test_devices,
    skip_if_gpu_unavailable,
)

import muGrid

# =============================================================================
# Test Configuration
# =============================================================================

GPU_AVAILABLE = muGrid.has_gpu


# =============================================================================
# 2D FEM Gradient Smoke Tests (Parametrized on Device)
# =============================================================================


@pytest.mark.parametrize("device", get_test_devices())
class TestFEMGradientOperator2D_smoke:
    """Smoke tests for FEMGradientOperator 2D on different devices."""

    def setup_method(self):
        self.nb_x_pts = 8
        self.nb_y_pts = 6

    def test_fem_gradient_2d_apply_smoke(self, device):
        """Test that FEM gradient 2D can be applied on device fields."""
        skip_if_gpu_unavailable(device)

        fem_grad = muGrid.FEMGradientOperator(2)  # 2D

        device_obj = create_device(device)
        fc_kwargs = {"device": device_obj} if device_obj else {}

        # Create collection with ghosts
        fc = muGrid.GlobalFieldCollection(
            (self.nb_x_pts, self.nb_y_pts),
            sub_pts={"quad": fem_grad.nb_quad_pts},
            nb_ghosts_right=(1, 1),
            **fc_kwargs,
        )

        # Create nodal and gradient fields
        nodal = fc.real_field("nodal", (1,))
        gradient = fc.real_field("gradient", (fem_grad.nb_output_components,), "quad")

        # Initialize and apply (should not raise)
        nodal.set_zero()
        gradient.set_zero()
        fem_grad.apply(nodal, gradient)

        # Verify device location
        if device == "gpu":
            assert gradient.is_on_gpu

    def test_fem_gradient_2d_transpose_smoke(self, device):
        """Test that FEM gradient 2D transpose can be applied on device fields."""
        skip_if_gpu_unavailable(device)

        fem_grad = muGrid.FEMGradientOperator(2)

        device_obj = create_device(device)
        fc_kwargs = {"device": device_obj} if device_obj else {}

        fc = muGrid.GlobalFieldCollection(
            (self.nb_x_pts, self.nb_y_pts),
            sub_pts={"quad": fem_grad.nb_quad_pts},
            nb_ghosts_right=(1, 1),
            **fc_kwargs,
        )

        nodal = fc.real_field("nodal", (1,))
        gradient = fc.real_field("gradient", (fem_grad.nb_output_components,), "quad")

        gradient.set_zero()
        nodal.set_zero()
        fem_grad.transpose(gradient, nodal)

        # Verify device location
        if device == "gpu":
            assert nodal.is_on_gpu


# =============================================================================
# 3D FEM Gradient Smoke Tests (Parametrized on Device)
# =============================================================================


@pytest.mark.parametrize("device", get_test_devices())
class TestFEMGradientOperator3D_smoke:
    """Smoke tests for 3D FEMGradientOperator on different devices."""

    def setup_method(self):
        self.nb_x_pts = 6
        self.nb_y_pts = 6
        self.nb_z_pts = 6

    def test_fem_gradient_3d_apply_smoke(self, device):
        """Test that FEM gradient 3D can be applied on device fields."""
        skip_if_gpu_unavailable(device)

        fem_grad = muGrid.FEMGradientOperator(3)  # 3D

        device_obj = create_device(device)
        fc_kwargs = {"device": device_obj} if device_obj else {}

        fc = muGrid.GlobalFieldCollection(
            (self.nb_x_pts, self.nb_y_pts, self.nb_z_pts),
            sub_pts={"quad": fem_grad.nb_quad_pts},
            nb_ghosts_right=(1, 1, 1),
            **fc_kwargs,
        )

        nodal = fc.real_field("nodal", (1,))
        gradient = fc.real_field("gradient", (fem_grad.nb_output_components,), "quad")

        nodal.set_zero()
        gradient.set_zero()
        fem_grad.apply(nodal, gradient)

        # Verify device location
        if device == "gpu":
            assert gradient.is_on_gpu

    def test_fem_gradient_3d_transpose_smoke(self, device):
        """Test that FEM gradient 3D transpose can be applied on device fields."""
        skip_if_gpu_unavailable(device)

        fem_grad = muGrid.FEMGradientOperator(3)

        device_obj = create_device(device)
        fc_kwargs = {"device": device_obj} if device_obj else {}

        fc = muGrid.GlobalFieldCollection(
            (self.nb_x_pts, self.nb_y_pts, self.nb_z_pts),
            sub_pts={"quad": fem_grad.nb_quad_pts},
            nb_ghosts_right=(1, 1, 1),
            **fc_kwargs,
        )

        nodal = fc.real_field("nodal", (1,))
        gradient = fc.real_field("gradient", (fem_grad.nb_output_components,), "quad")

        gradient.set_zero()
        nodal.set_zero()
        fem_grad.transpose(gradient, nodal)

        # Verify device location
        if device == "gpu":
            assert nodal.is_on_gpu


# =============================================================================
# GPU vs CPU Correctness Tests (Dual-device comparison)
# =============================================================================


@pytest.mark.skipif(not GPU_AVAILABLE, reason="GPU backend not available")
class TestFEMGradientOperatorGPUCorrectness:
    """Test suite for FEMGradientOperator GPU correctness.

    These tests verify that GPU FEM gradient operations return CuPy arrays
    and produce correct results that match the CPU reference.
    """

    def setup_method(self, method):
        """Skip tests if no GPU is available at runtime."""
        if not muGrid.is_gpu_available():
            pytest.skip("No GPU device available at runtime")
        if not HAS_CUPY:
            pytest.skip("CuPy not available for GPU tests")
        self.nb_x_pts = 8
        self.nb_y_pts = 6

    def test_device_fem_gradient_returns_cupy(self):
        """Test that FEM gradient on device fields returns CuPy arrays."""
        fem_grad = muGrid.FEMGradientOperator(2)

        fc = muGrid.GlobalFieldCollection(
            (self.nb_x_pts, self.nb_y_pts),
            sub_pts={"quad": fem_grad.nb_quad_pts},
            nb_ghosts_right=(1, 1),
            device=muGrid.Device.gpu(),
        )

        nodal = fc.real_field("nodal", (1,))
        gradient = fc.real_field("gradient", (fem_grad.nb_output_components,), "quad")

        # Arrays should be CuPy
        assert isinstance(nodal.s, cp.ndarray)
        assert isinstance(gradient.s, cp.ndarray)

        fem_grad.apply(nodal, gradient)

        # Result should still be CuPy
        assert isinstance(gradient.s, cp.ndarray)

    def test_device_fem_gradient_2d_correctness(self):
        """Test that 2D device FEM gradient produces correct results.

        Compare GPU results against CPU reference.
        """
        fem_grad = muGrid.FEMGradientOperator(2)

        # Create host reference
        fc_host = muGrid.GlobalFieldCollection(
            (self.nb_x_pts, self.nb_y_pts),
            sub_pts={"quad": fem_grad.nb_quad_pts},
            nb_ghosts_right=(1, 1),
        )
        nodal_host = fc_host.real_field("nodal", (1,))
        gradient_host = fc_host.real_field(
            "gradient", (fem_grad.nb_output_components,), "quad"
        )

        # Create device fields
        fc_device = muGrid.GlobalFieldCollection(
            (self.nb_x_pts, self.nb_y_pts),
            sub_pts={"quad": fem_grad.nb_quad_pts},
            nb_ghosts_right=(1, 1),
            device=muGrid.Device.gpu(),
        )
        nodal_device = fc_device.real_field("nodal", (1,))
        gradient_device = fc_device.real_field(
            "gradient", (fem_grad.nb_output_components,), "quad"
        )

        # Initialize with same random data
        test_data = np.random.rand(*nodal_host.pg.shape)
        nodal_host.pg[...] = test_data
        nodal_device.pg[...] = cp.asarray(test_data)

        # Apply on both
        fem_grad.apply(nodal_host, gradient_host)
        fem_grad.apply(nodal_device, gradient_device)

        # Compare results
        host_result = gradient_host.s
        device_result = cp.asnumpy(gradient_device.s)

        np.testing.assert_allclose(device_result, host_result, rtol=1e-10)

    def test_device_fem_gradient_2d_transpose_correctness(self):
        """Test that 2D device FEM transpose produces correct results."""
        fem_grad = muGrid.FEMGradientOperator(2)

        # Create host reference
        fc_host = muGrid.GlobalFieldCollection(
            (self.nb_x_pts, self.nb_y_pts),
            sub_pts={"quad": fem_grad.nb_quad_pts},
            nb_ghosts_right=(1, 1),
        )
        nodal_host = fc_host.real_field("nodal", (1,))
        gradient_host = fc_host.real_field(
            "gradient", (fem_grad.nb_output_components,), "quad"
        )

        # Create device fields
        fc_device = muGrid.GlobalFieldCollection(
            (self.nb_x_pts, self.nb_y_pts),
            sub_pts={"quad": fem_grad.nb_quad_pts},
            nb_ghosts_right=(1, 1),
            device=muGrid.Device.gpu(),
        )
        nodal_device = fc_device.real_field("nodal", (1,))
        gradient_device = fc_device.real_field(
            "gradient", (fem_grad.nb_output_components,), "quad"
        )

        # Initialize gradient with random data
        test_data = np.random.rand(*gradient_host.s.shape)
        gradient_host.s[...] = test_data
        gradient_device.s[...] = cp.asarray(test_data)

        # Apply transpose on both
        fem_grad.transpose(gradient_host, nodal_host)
        fem_grad.transpose(gradient_device, nodal_device)

        # Compare results
        host_result = nodal_host.s
        device_result = cp.asnumpy(nodal_device.s)

        np.testing.assert_allclose(device_result, host_result, rtol=1e-10)

    def test_device_fem_gradient_3d_correctness(self):
        """Test that 3D device FEM gradient produces correct results."""
        nb_x, nb_y, nb_z = 6, 6, 6
        fem_grad = muGrid.FEMGradientOperator(3)

        # Create host reference
        fc_host = muGrid.GlobalFieldCollection(
            (nb_x, nb_y, nb_z),
            sub_pts={"quad": fem_grad.nb_quad_pts},
            nb_ghosts_right=(1, 1, 1),
        )
        nodal_host = fc_host.real_field("nodal", (1,))
        gradient_host = fc_host.real_field(
            "gradient", (fem_grad.nb_output_components,), "quad"
        )

        # Create device fields
        fc_device = muGrid.GlobalFieldCollection(
            (nb_x, nb_y, nb_z),
            sub_pts={"quad": fem_grad.nb_quad_pts},
            nb_ghosts_right=(1, 1, 1),
            device=muGrid.Device.gpu(),
        )
        nodal_device = fc_device.real_field("nodal", (1,))
        gradient_device = fc_device.real_field(
            "gradient", (fem_grad.nb_output_components,), "quad"
        )

        # Initialize with same random data
        test_data = np.random.rand(*nodal_host.pg.shape)
        nodal_host.pg[...] = test_data
        nodal_device.pg[...] = cp.asarray(test_data)

        # Apply on both
        fem_grad.apply(nodal_host, gradient_host)
        fem_grad.apply(nodal_device, gradient_device)

        # Compare results
        host_result = gradient_host.s
        device_result = cp.asnumpy(gradient_device.s)

        np.testing.assert_allclose(device_result, host_result, rtol=1e-10)

    def test_device_fem_gradient_3d_transpose_correctness(self):
        """Test that 3D device FEM transpose produces correct results."""
        nb_x, nb_y, nb_z = 6, 6, 6
        fem_grad = muGrid.FEMGradientOperator(3)

        # Create host reference
        fc_host = muGrid.GlobalFieldCollection(
            (nb_x, nb_y, nb_z),
            sub_pts={"quad": fem_grad.nb_quad_pts},
            nb_ghosts_right=(1, 1, 1),
        )
        nodal_host = fc_host.real_field("nodal", (1,))
        gradient_host = fc_host.real_field(
            "gradient", (fem_grad.nb_output_components,), "quad"
        )

        # Create device fields
        fc_device = muGrid.GlobalFieldCollection(
            (nb_x, nb_y, nb_z),
            sub_pts={"quad": fem_grad.nb_quad_pts},
            nb_ghosts_right=(1, 1, 1),
            device=muGrid.Device.gpu(),
        )
        nodal_device = fc_device.real_field("nodal", (1,))
        gradient_device = fc_device.real_field(
            "gradient", (fem_grad.nb_output_components,), "quad"
        )

        # Initialize gradient with random data
        test_data = np.random.rand(*gradient_host.s.shape)
        gradient_host.s[...] = test_data
        gradient_device.s[...] = cp.asarray(test_data)

        # Apply transpose on both
        fem_grad.transpose(gradient_host, nodal_host)
        fem_grad.transpose(gradient_device, nodal_device)

        # Compare results
        host_result = nodal_host.s
        device_result = cp.asnumpy(nodal_device.s)

        np.testing.assert_allclose(device_result, host_result, rtol=1e-10)

    def test_device_fem_gradient_roundtrip(self):
        """Test round-trip: B^T * B should be symmetric positive semi-definite."""
        fem_grad = muGrid.FEMGradientOperator(2)

        # Create device collection
        fc = muGrid.GlobalFieldCollection(
            (self.nb_x_pts, self.nb_y_pts),
            sub_pts={"quad": fem_grad.nb_quad_pts},
            nb_ghosts_right=(1, 1),
            device=muGrid.Device.gpu(),
        )

        nodal_in = fc.real_field("nodal_in", (1,))
        gradient = fc.real_field("gradient", (fem_grad.nb_output_components,), "quad")
        nodal_out = fc.real_field("nodal_out", (1,))

        # Initialize with random data
        test_data = cp.random.rand(*nodal_in.pg.shape)
        nodal_in.pg[...] = test_data

        # Apply B then B^T
        fem_grad.apply(nodal_in, gradient)
        fem_grad.transpose(gradient, nodal_out)

        # Result should be a valid array (no NaN or Inf)
        result = cp.asnumpy(nodal_out.s)
        assert not np.any(np.isnan(result))
        assert not np.any(np.isinf(result))


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



import os

import numpy as np
import pytest
from NuMPI import MPI
from NuMPI.Testing.Assertions import assert_all_allclose
from NuMPI.Testing.Subdivision import suggest_subdivisions

import muGrid
from muGrid import GlobalFieldCollection

try:
    import netCDF4
    HAS_NETCDF4 = True
except ImportError:
    HAS_NETCDF4 = False

# Check if muGrid was built with NetCDF support
HAS_MUGRID_NETCDF = hasattr(muGrid, 'OpenMode')


def get_nb_subdivisions(nb_processes: int):
    subdivision_setup = {
        1: [(1,), (1, 1)],
        2: [
            (2,),
        ],
        4: [
            (4,),
            (2, 2),
        ],
        8: [
            (8,),
            (2, 2, 2),
            (4, 2, 1),
            (8, 1, 1),
        ],
    }
    if nb_processes in subdivision_setup:
        return subdivision_setup[nb_processes]
    else:
        raise NotImplementedError("Not planned for this number of processes.")


def make_subdivisions():
    try:
        from mpi4py import MPI

        comm = muGrid.Communicator(MPI.COMM_WORLD)
    except ImportError:
        comm = muGrid.Communicator()
    nb_processes = comm.size

    # Create a Cartesian decomposition
    nb_subdivisions = get_nb_subdivisions(nb_processes)

    return [(comm, s) for s in nb_subdivisions]


@pytest.mark.parametrize("comm,nb_subdivisions", make_subdivisions())
def test_communicate_ghosts(comm, nb_subdivisions):
    # Create a Cartesian decomposition
    spatial_dim = len(nb_subdivisions)
    nb_pts_per_dim = 5
    nb_domain_grid_pts = np.full(spatial_dim, nb_pts_per_dim)
    nb_ghosts_left = np.full(spatial_dim, 2)
    nb_ghosts_right = np.full(spatial_dim, 2)
    cart_decomp = muGrid.CartesianDecomposition(
        comm,
        nb_domain_grid_pts.tolist(),
        nb_subdivisions,
        nb_ghosts_left.tolist(),
        nb_ghosts_right.tolist(),
    )

    # Idiot check the subdivision
    for i in range(len(nb_subdivisions)):
        s = comm.sum(cart_decomp.nb_subdomain_grid_pts[i])
        assert (
            s
            == cart_decomp.nb_domain_grid_pts[i]
            * np.prod(cart_decomp.nb_subdivisions)
            / cart_decomp.nb_subdivisions[i]
        )

    # Create a field for testing
    field_name = "test_field"
    field = cart_decomp.real_field(field_name)

    # Create reference values
    global_coords = cart_decomp.icoordsg
    weights = np.arange(spatial_dim) + 1
    ref_values = np.einsum("i, i...->...", weights, global_coords)

    # Fill the field, non-ghost with reference values, ghost with some other value
    nb_subdomain_grid_pts = cart_decomp.nb_subdomain_grid_pts_with_ghosts
    for index in np.ndindex(*nb_subdomain_grid_pts):
        is_not_ghost = all(
            idx >= nb_ghosts_left[dim]
            and idx < nb_subdomain_grid_pts[dim] - nb_ghosts_right[dim]
            for dim, idx in enumerate(index)
        )
        if is_not_ghost:
            field.sg[(..., *index)] = ref_values[(..., *index)]
        else:
            field.sg[(..., *index)] = -1

    # Check accessors
    np.testing.assert_array_equal(
        field.s.shape[-spatial_dim:],
        np.array(field.sg.shape)[-spatial_dim:] - nb_ghosts_left - nb_ghosts_right,
    )
    np.testing.assert_array_equal(
        field.p.shape[-spatial_dim:],
        np.array(field.pg.shape)[-spatial_dim:] - nb_ghosts_left - nb_ghosts_right,
    )

    # Communicate ghost cells
    cart_decomp.communicate_ghosts(field)

    # Check values at all grid points
    for index in np.ndindex(*nb_subdomain_grid_pts):
        np.testing.assert_allclose(
            field.sg[(..., *index)],
            ref_values[(..., *index)],
            err_msg=f"Mismatch at {index}",
        )


def test_field_accessors(comm, nb_grid_pts=(128, 128)):
    s = suggest_subdivisions(len(nb_grid_pts), comm.size)

    decomposition = muGrid.CartesianDecomposition(comm, nb_grid_pts, s, (1, 1), (1, 1))

    field = decomposition.real_field("test-field")

    xg, yg = decomposition.coordsg
    field.pg[...] = xg + 100 * yg

    assert_all_allclose(MPI.COMM_WORLD, field.pg[..., 1:-1, 1:-1], field.p)
    assert_all_allclose(MPI.COMM_WORLD, field.sg[..., 1:-1, 1:-1], field.s)

    # Test in-place assignment
    field.pg[...] = np.random.random(field.pg.shape)

    assert_all_allclose(MPI.COMM_WORLD, field.pg[..., 1:-1, 1:-1], field.p)
    assert_all_allclose(MPI.COMM_WORLD, field.sg[..., 1:-1, 1:-1], field.s)


@pytest.mark.skipif(not HAS_MUGRID_NETCDF, reason="muGrid built without NetCDF support")
@pytest.mark.parametrize("comm,nb_subdivisions", make_subdivisions())
def test_io(comm, nb_subdivisions):
    filename = "test_io_output.nc"

    # Create a Cartesian decomposition
    spatial_dim = len(nb_subdivisions)
    nb_pts_per_dim = 5
    nb_domain_grid_pts = np.full(spatial_dim, nb_pts_per_dim)
    nb_ghost_left = np.full(spatial_dim, 1)
    nb_ghost_right = np.full(spatial_dim, 2)
    cart_decomp = muGrid.CartesianDecomposition(
        comm,
        nb_domain_grid_pts.tolist(),
        nb_subdivisions,
        nb_ghost_left.tolist(),
        nb_ghost_right.tolist(),
    )

    # Create a field for testing
    field_name = "test_field"
    field = cart_decomp.real_field(field_name)

    field.pg[...] = (cart_decomp.icoordsg**2).sum(axis=0)

    # Write to file
    try:
        f = muGrid.FileIONetCDF(filename, muGrid.OpenMode.Overwrite, comm)
    except RuntimeError as e:
        print(f"Opening file for write failed on rank {comm.rank}/{comm.size}")
        raise e
    f.register_field_collection(cart_decomp)
    f.append_frame().write()
    f.close()


@pytest.mark.parametrize("comm,nb_subdivisions", make_subdivisions())
def test_reduce_ghosts(comm, nb_subdivisions):
    """Test that reduce_ghosts correctly accumulates ghost contributions to interior."""
    # Create a Cartesian decomposition
    spatial_dim = len(nb_subdivisions)
    nb_pts_per_dim = 5
    nb_domain_grid_pts = np.full(spatial_dim, nb_pts_per_dim)
    nb_ghosts_left = np.full(spatial_dim, 2)
    nb_ghosts_right = np.full(spatial_dim, 2)
    cart_decomp = muGrid.CartesianDecomposition(
        comm,
        nb_domain_grid_pts.tolist(),
        nb_subdivisions,
        nb_ghosts_left.tolist(),
        nb_ghosts_right.tolist(),
    )

    # Create a field for testing
    field = cart_decomp.real_field("reduce_test")

    # Get subdomain dimensions
    nb_subdomain_grid_pts = cart_decomp.nb_subdomain_grid_pts_with_ghosts

    # Fill interior with ones, ghosts with a pattern that should accumulate
    # After reduce_ghosts:
    # - Interior boundary points should have accumulated contributions
    # - Ghost regions should be zeroed
    for index in np.ndindex(*nb_subdomain_grid_pts):
        is_interior = all(
            idx >= nb_ghosts_left[dim]
            and idx < nb_subdomain_grid_pts[dim] - nb_ghosts_right[dim]
            for dim, idx in enumerate(index)
        )
        if is_interior:
            field.sg[(..., *index)] = 1.0
        else:
            # Ghost points get value 2.0 - these should accumulate to boundary interior
            field.sg[(..., *index)] = 2.0

    # Store original interior values for comparison
    interior_before = field.s.copy()

    # Perform ghost reduction
    cart_decomp.reduce_ghosts(field)

    # Check that ghost regions are now zero
    for index in np.ndindex(*nb_subdomain_grid_pts):
        is_ghost = any(
            idx < nb_ghosts_left[dim]
            or idx >= nb_subdomain_grid_pts[dim] - nb_ghosts_right[dim]
            for dim, idx in enumerate(index)
        )
        if is_ghost:
            np.testing.assert_array_equal(
                field.sg[(..., *index)],
                0.0,
                err_msg=f"Ghost region at {index} not zeroed after reduce",
            )

    # For single-process periodic case, boundary interior points should have
    # accumulated ghost contributions
    if comm.size == 1:
        # The accumulated values depend on the ghost structure
        # At minimum, interior values should be >= original values
        assert np.all(field.s >= interior_before), \
            "Interior values should not decrease after reduction"


@pytest.mark.parametrize("comm,nb_subdivisions", make_subdivisions())
def test_reduce_ghosts_multicomponent(comm, nb_subdivisions):
    """Test reduce_ghosts with multi-component fields."""
    spatial_dim = len(nb_subdivisions)
    nb_pts_per_dim = 4
    nb_domain_grid_pts = np.full(spatial_dim, nb_pts_per_dim)
    nb_ghosts_left = np.full(spatial_dim, 1)
    nb_ghosts_right = np.full(spatial_dim, 1)

    cart_decomp = muGrid.CartesianDecomposition(
        comm,
        nb_domain_grid_pts.tolist(),
        nb_subdivisions,
        nb_ghosts_left.tolist(),
        nb_ghosts_right.tolist(),
    )

    # Create a multi-component field (e.g., vector field)
    nb_components = 3
    field = cart_decomp.real_field("vector_field", nb_components)

    # Fill with component-dependent values
    for comp in range(nb_components):
        field.sg[comp, ...] = comp + 1.0

    # Reduce ghosts
    cart_decomp.reduce_ghosts(field)

    # Check ghost regions are zeroed for all components using slicing
    # Build slices for ghost regions
    # Note: field.sg has shape (nb_components, nb_sub_pts, *spatial_dims)
    nb_subdomain_grid_pts = cart_decomp.nb_subdomain_grid_pts_with_ghosts
    nb_prefix_dims = len(field.sg.shape) - spatial_dim  # components + sub_pts

    # Check left ghost regions
    for dim in range(spatial_dim):
        if nb_ghosts_left[dim] > 0:
            slices = [slice(None)] * len(field.sg.shape)
            slices[nb_prefix_dims + dim] = slice(0, nb_ghosts_left[dim])
            np.testing.assert_array_equal(
                field.sg[tuple(slices)],
                0.0,
                err_msg=f"Left ghost region in dim {dim} not zeroed",
            )

    # Check right ghost regions
    for dim in range(spatial_dim):
        if nb_ghosts_right[dim] > 0:
            slices = [slice(None)] * len(field.sg.shape)
            slices[nb_prefix_dims + dim] = slice(
                nb_subdomain_grid_pts[dim] - nb_ghosts_right[dim], None
            )
            np.testing.assert_array_equal(
                field.sg[tuple(slices)],
                0.0,
                err_msg=f"Right ghost region in dim {dim} not zeroed",
            )


@pytest.mark.parametrize("comm,nb_subdivisions", make_subdivisions())
def test_reduce_ghosts_asymmetric(comm, nb_subdivisions):
    """Test reduce_ghosts with asymmetric ghost buffer sizes."""
    spatial_dim = len(nb_subdivisions)
    nb_pts_per_dim = 6
    nb_domain_grid_pts = np.full(spatial_dim, nb_pts_per_dim)
    # Asymmetric: more ghosts on right than left
    nb_ghosts_left = np.full(spatial_dim, 1)
    nb_ghosts_right = np.full(spatial_dim, 2)

    cart_decomp = muGrid.CartesianDecomposition(
        comm,
        nb_domain_grid_pts.tolist(),
        nb_subdivisions,
        nb_ghosts_left.tolist(),
        nb_ghosts_right.tolist(),
    )

    field = cart_decomp.real_field("asymmetric_test")

    # Fill with pattern: interior=1, left_ghost=10, right_ghost=20
    nb_subdomain_grid_pts = cart_decomp.nb_subdomain_grid_pts_with_ghosts
    for index in np.ndindex(*nb_subdomain_grid_pts):
        is_left_ghost = any(idx < nb_ghosts_left[dim] for dim, idx in enumerate(index))
        is_right_ghost = any(
            idx >= nb_subdomain_grid_pts[dim] - nb_ghosts_right[dim]
            for dim, idx in enumerate(index)
        )
        if is_left_ghost:
            field.sg[(..., *index)] = 10.0
        elif is_right_ghost:
            field.sg[(..., *index)] = 20.0
        else:
            field.sg[(..., *index)] = 1.0

    # Reduce
    cart_decomp.reduce_ghosts(field)

    # Verify ghosts are zeroed
    for index in np.ndindex(*nb_subdomain_grid_pts):
        is_ghost = any(
            idx < nb_ghosts_left[dim]
            or idx >= nb_subdomain_grid_pts[dim] - nb_ghosts_right[dim]
            for dim, idx in enumerate(index)
        )
        if is_ghost:
            np.testing.assert_array_equal(
                field.sg[(..., *index)],
                0.0,
                err_msg=f"Ghost at {index} not zeroed (asymmetric case)",
            )


@pytest.mark.parametrize("comm,nb_subdivisions", make_subdivisions())
def test_reduce_ghosts_adjoint_property(comm, nb_subdivisions):
    """
    Test that reduce_ghosts is the adjoint of communicate_ghosts.

    For periodic BC, if we define:
    - C = communicate_ghosts (fills ghosts from interior)
    - R = reduce_ghosts (accumulates ghosts to interior, zeros ghosts)

    Then R should be the adjoint of C in the sense that:
    <C(x), y> = <x, R(y)>

    where <.,.> is the inner product over the full domain (including ghosts).
    """
    spatial_dim = len(nb_subdivisions)
    # Use enough points to ensure every rank has at least 1 interior point
    # even with maximum subdivision (8 in any dimension for 8 processes)
    nb_pts_per_dim = max(8, max(nb_subdivisions))
    nb_domain_grid_pts = np.full(spatial_dim, nb_pts_per_dim)
    nb_ghosts_left = np.full(spatial_dim, 1)
    nb_ghosts_right = np.full(spatial_dim, 1)

    cart_decomp = muGrid.CartesianDecomposition(
        comm,
        nb_domain_grid_pts.tolist(),
        nb_subdivisions,
        nb_ghosts_left.tolist(),
        nb_ghosts_right.tolist(),
    )

    # Create two fields
    field_x = cart_decomp.real_field("x")
    field_y = cart_decomp.real_field("y")

    # Initialize x with random interior values and zero ghosts
    # Initialize y with random values everywhere (interior + ghosts)
    np.random.seed(42 + comm.rank)
    field_x.sg[...] = np.random.rand(*field_x.sg.shape)
    field_y.sg[...] = np.random.rand(*field_y.sg.shape)

    # Zero out x's ghosts (x represents interior-only data)
    nb_subdomain_grid_pts = cart_decomp.nb_subdomain_grid_pts_with_ghosts
    for index in np.ndindex(*nb_subdomain_grid_pts):
        is_ghost = any(
            idx < nb_ghosts_left[dim]
            or idx >= nb_subdomain_grid_pts[dim] - nb_ghosts_right[dim]
            for dim, idx in enumerate(index)
        )
        if is_ghost:
            field_x.sg[(..., *index)] = 0.0

    # Store original values
    x_original = field_x.sg.copy()  # x with interior values, zero ghosts
    y_original = field_y.sg.copy()  # y with values everywhere

    # Compute C(x) = communicate_ghosts(x)
    cart_decomp.communicate_ghosts(field_x)

    # <C(x), y> - inner product over full ghosted domain
    inner_Cx_y = comm.sum(np.sum(field_x.sg * y_original))

    # Compute R(y) = reduce_ghosts(y)
    # Restore y to original values first
    field_y.sg[...] = y_original
    cart_decomp.reduce_ghosts(field_y)

    # <x, R(y)> - inner product with original x (zero ghosts)
    inner_x_Ry = comm.sum(np.sum(x_original * field_y.sg))

    # The adjoint property: <C(x), y> = <x, R(y)>
    np.testing.assert_allclose(
        inner_Cx_y,
        inner_x_Ry,
        rtol=1e-10,
        err_msg="reduce_ghosts is not adjoint of communicate_ghosts",
    )


@pytest.mark.parametrize("comm,nb_subdivisions", make_subdivisions())
def test_large_ghost_buffers(comm, nb_subdivisions):
    # Create a Cartesian decomposition
    spatial_dim = len(nb_subdivisions)
    nb_pts_per_dim = 5
    nb_domain_grid_pts = np.full(spatial_dim, nb_pts_per_dim)
    nb_ghosts_left = np.full(spatial_dim, 5)
    nb_ghosts_right = np.full(spatial_dim, 5)
    muGrid.CartesianDecomposition(
        comm,
        nb_domain_grid_pts.tolist(),
        nb_subdivisions,
        nb_ghosts_left.tolist(),
        nb_ghosts_right.tolist(),
    )


@pytest.mark.skipif(not HAS_MUGRID_NETCDF, reason="muGrid built without NetCDF support")
@pytest.mark.skipif(not HAS_NETCDF4, reason="netCDF4 not available")
@pytest.mark.parametrize("comm,nb_subdivisions", make_subdivisions())
def test_fileio_netcdf_ghost_offset(comm, nb_subdivisions):
    """Test that FileIONetCDF writes interior data, not ghost-shifted data."""
    spatial_dim = len(nb_subdivisions)
    nb_domain_grid_pts = np.full(spatial_dim, 4)
    nb_ghost_left = np.full(spatial_dim, 1)
    nb_ghost_right = np.full(spatial_dim, 1)

    cart_decomp = muGrid.CartesianDecomposition(
        comm,
        nb_domain_grid_pts.tolist(),
        nb_subdivisions,
        nb_ghost_left.tolist(),
        nb_ghost_right.tolist(),
    )

    field = cart_decomp.real_field("test_field")

    # Fill pg with weighted coordinate sums to make shifts detectable
    global_coords = cart_decomp.icoordsg
    weights = np.array([100**i for i in range(spatial_dim)])
    field.pg[...] = np.einsum("i, i...->...", weights, global_coords)

    expected_interior = field.p.copy()

    filename = "test_ghost_offset.nc"

    try:
        file_io = muGrid.FileIONetCDF(
            filename, muGrid.OpenMode.Overwrite, communicator=comm
        )
        file_io.register_field_collection(
            cart_decomp, field_names=["test_field"]
        )
        file_io.append_frame().write()
        file_io.close()

        comm.barrier()

        # Each rank reads full file and checks its own slice
        with netCDF4.Dataset(filename, "r") as nc:
            stored_data = np.asarray(nc.variables["test_field"][0])

        # Build slice for this rank's subdomain in global array
        slices = tuple(
            slice(loc, loc + size)
            for loc, size in zip(
                cart_decomp.subdomain_locations,
                cart_decomp.nb_subdomain_grid_pts,
            )
        )

        np.testing.assert_array_equal(
            stored_data[slices],
            expected_interior
        )
    finally:
        comm.barrier()
        if comm.rank == 0 and os.path.exists(filename):
            os.unlink(filename)


@pytest.mark.parametrize("comm,nb_subdivisions", make_subdivisions())
def test_collection_is_wrapped(comm, nb_subdivisions):
    """
    Test that CartesianDecomposition.collection returns a wrapped
    GlobalFieldCollection.
    """
    spatial_dim = len(nb_subdivisions)
    nb_pts_per_dim = 5
    nb_domain_grid_pts = np.full(spatial_dim, nb_pts_per_dim)
    nb_ghosts_left = np.full(spatial_dim, 1)
    nb_ghosts_right = np.full(spatial_dim, 1)

    cart_decomp = muGrid.CartesianDecomposition(
        comm,
        nb_domain_grid_pts.tolist(),
        nb_subdivisions,
        nb_ghosts_left.tolist(),
        nb_ghosts_right.tolist(),
    )

    collection = cart_decomp.collection

    # Check that it's the wrapped type, not the raw C++ type
    assert isinstance(collection, GlobalFieldCollection)
    # Check that the wrapper has the expected attributes
    assert hasattr(collection, "_cpp")
    assert hasattr(collection, "nb_grid_pts")


@pytest.mark.parametrize("comm,nb_subdivisions", make_subdivisions())
def test_collection_field_creation(comm, nb_subdivisions):
    """Test that fields created via collection property work correctly."""
    spatial_dim = len(nb_subdivisions)
    nb_pts_per_dim = 5
    nb_domain_grid_pts = np.full(spatial_dim, nb_pts_per_dim)
    nb_ghosts_left = np.full(spatial_dim, 1)
    nb_ghosts_right = np.full(spatial_dim, 1)

    cart_decomp = muGrid.CartesianDecomposition(
        comm,
        nb_domain_grid_pts.tolist(),
        nb_subdivisions,
        nb_ghosts_left.tolist(),
        nb_ghosts_right.tolist(),
    )

    # Create field via collection property
    collection = cart_decomp.collection
    field = collection.real_field("test_via_collection")

    # Check that field works correctly
    assert field is not None
    # Fill with data
    field.sg[...] = 1.0

    # Communicate ghosts should work with field from collection
    cart_decomp.communicate_ghosts(field)

    # Verify ghost values are properly filled (all ones for constant field)
    np.testing.assert_array_almost_equal(field.sg, 1.0)



"""
Pytest configuration and shared fixtures for muGrid tests.

This module provides:
- Common fixtures (comm, gpu_device)
- GPU testing utilities (skip_if_gpu_unavailable, get_test_devices, create_device)
- CuPy availability checking
"""

import pytest

import muGrid

# =============================================================================
# CuPy availability
# =============================================================================

try:
    import cupy as cp

    HAS_CUPY = True
except ImportError:
    HAS_CUPY = False
    cp = None


# =============================================================================
# GPU testing utilities
# =============================================================================


def get_test_devices():
    """Return list of device strings for parametrized tests.

    Returns ["cpu"] if GPU support is not compiled in, otherwise ["cpu", "gpu"].
    Use with @pytest.mark.parametrize("device", get_test_devices()).
    """
    devices = ["cpu"]
    if muGrid.has_gpu:
        devices.append("gpu")
    return devices


def create_device(device_str):
    """Create a Device object from string.

    Parameters
    ----------
    device_str : str
        Either "cpu" or "gpu".

    Returns
    -------
    Device or None
        None for CPU (uses default), Device.gpu() for GPU.
    """
    if device_str == "cpu":
        return None  # CPU uses default (no device argument)
    elif device_str == "gpu":
        return muGrid.Device.gpu()
    raise ValueError(f"Unknown device: {device_str}")


def skip_if_gpu_unavailable(device):
    """Skip a test if GPU is requested but not available.

    Call this at the start of parametrized tests that use the "device" parameter.
    Skips if:
    - device is "gpu" and no GPU hardware is available at runtime
    - device is "gpu" and CuPy is not installed (needed for GPU array access)

    Parameters
    ----------
    device : str
        The device string ("cpu" or "gpu").

    Example
    -------
    @pytest.mark.parametrize("device", get_test_devices())
    def test_something(device):
        skip_if_gpu_unavailable(device)
        # ... rest of test
    """
    if device == "gpu":
        if not muGrid.is_gpu_available():
            pytest.skip("No GPU device available at runtime")
        if not HAS_CUPY:
            pytest.skip("CuPy not available for GPU array access")


def get_array_module(device):
    """Get the array module (numpy or cupy) for the given device.

    Parameters
    ----------
    device : str
        Either "cpu" or "gpu".

    Returns
    -------
    module
        numpy for CPU, cupy for GPU.
    """
    if device == "gpu":
        return cp
    import numpy as np

    return np


# =============================================================================
# Fixtures
# =============================================================================


@pytest.fixture
def comm():
    """Provide an MPI communicator (or serial fallback)."""
    try:
        from mpi4py import MPI

        return muGrid.Communicator(MPI.COMM_WORLD)
    except ImportError:
        return muGrid.Communicator()


@pytest.fixture
def gpu_device():
    """Provide a GPU device, skipping if unavailable.

    Use this fixture when a test requires GPU hardware.
    """
    if not muGrid.is_gpu_available():
        pytest.skip("No GPU device available at runtime")
    if not HAS_CUPY:
        pytest.skip("CuPy not available for GPU array access")
    return muGrid.Device.gpu()



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@file     python_field_gpu_tests.py

@author  Lars Pastewka <lars.pastewka@imtek.uni-freiburg.de>

@date    09 Dec 2024

@brief   Test GPU field functionality (CUDA/ROCm)

Copyright © 2024 Lars Pastewka

µGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

µGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with µGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

import unittest

import numpy as np

import muGrid

# Try to import CuPy for GPU tests
try:
    import cupy as cp
    HAS_CUPY = True
except ImportError:
    HAS_CUPY = False


# Use compile-time feature flag for GPU availability
GPU_AVAILABLE = muGrid.has_gpu


class HostFieldDeviceInfoTests(unittest.TestCase):
    """Test device introspection for host fields."""

    def setUp(self):
        self.nb_grid_pts = (8, 8)
        self.fc = muGrid.GlobalFieldCollection(self.nb_grid_pts)

    def test_host_field_device_property(self):
        """Test that host fields report 'cpu' as device."""
        field_cpp = self.fc.register_real_field("test", 3)
        self.assertEqual(field_cpp.device, "cpu")

    def test_host_field_is_on_gpu(self):
        """Test that host fields report is_on_gpu = False."""
        field_cpp = self.fc.register_real_field("test", 3)
        self.assertFalse(field_cpp.is_on_gpu)

    def test_host_field_dlpack_device(self):
        """Test that host fields report correct DLPack device tuple."""
        field_cpp = self.fc.register_real_field("test", 3)
        device_type, device_id = field_cpp.__dlpack_device__()
        # kDLCPU = 1
        self.assertEqual(device_type, 1)
        self.assertEqual(device_id, 0)

    def test_wrapped_field_device_properties(self):
        """Test device properties through Python Field wrapper."""
        field = self.fc.real_field("wrapped_test", (3,))
        self.assertEqual(field.device, "cpu")
        self.assertFalse(field.is_on_gpu)

    def test_host_collection_is_on_device(self):
        """Test that host collection reports is_on_device = False."""
        self.assertFalse(self.fc.is_on_device)

    def test_host_collection_device(self):
        """Test that host collection reports CPU device."""
        self.assertTrue(self.fc.device.is_host)


class HostFieldAccessTests(unittest.TestCase):
    """Test host field memory access."""

    def setUp(self):
        self.nb_grid_pts = (8, 8)
        self.fc = muGrid.GlobalFieldCollection(self.nb_grid_pts)

    def test_host_field_returns_numpy(self):
        """Test that host field accessors return numpy arrays."""
        field = self.fc.real_field("numpy_test", (3,))
        arr = field.s
        self.assertIsInstance(arr, np.ndarray)

    def test_all_field_types_host(self):
        """Test all field types with host memory space."""
        real_f = self.fc.real_field("real_host")
        int_f = self.fc.int_field("int_host")
        uint_f = self.fc.uint_field("uint_host")
        complex_f = self.fc.complex_field("complex_host")

        for f in [real_f, int_f, uint_f, complex_f]:
            self.assertFalse(f.is_on_gpu)
            self.assertEqual(f.device, "cpu")


@unittest.skipUnless(GPU_AVAILABLE, "GPU backend not available")
class DeviceCollectionTests(unittest.TestCase):
    """Test device collection creation (requires GPU backend)."""

    def setUp(self):
        self.nb_grid_pts = (8, 8)
        # Create a device collection
        self.fc = muGrid.GlobalFieldCollection(
            self.nb_grid_pts,
            device=muGrid.Device.cuda()
        )

    def test_device_collection_is_on_device(self):
        """Test that device collection reports is_on_device = True."""
        self.assertTrue(self.fc.is_on_device)

    def test_device_collection_device(self):
        """Test that device collection reports CUDA/ROCm device."""
        self.assertTrue(self.fc.device.is_device)

    def test_device_field_is_on_gpu(self):
        """Test that fields in device collection report is_on_gpu = True."""
        field = self.fc.register_real_field("device_test", 3)
        self.assertTrue(field.is_on_gpu)

    def test_device_field_device_property(self):
        """Test that device fields report correct device string."""
        field = self.fc.register_real_field("device_test", 3)
        # Should be "cuda:0" or "rocm:0"
        device = field.device
        self.assertTrue(device.startswith("cuda:") or device.startswith("rocm:"))

    def test_device_field_dlpack_device(self):
        """Test that device fields report correct DLPack device tuple."""
        field = self.fc.register_real_field("device_test", 3)
        device_type, device_id = field.__dlpack_device__()
        # kDLCUDA = 2, kDLROCm = 10
        self.assertIn(device_type, [2, 10])
        self.assertEqual(device_id, 0)

    def test_all_device_field_types(self):
        """Test creation of all device field types."""
        real_f = self.fc.register_real_field("real_device", 2)
        complex_f = self.fc.register_complex_field("complex_device", 2)
        int_f = self.fc.register_int_field("int_device", 2)
        uint_f = self.fc.register_uint_field("uint_device", 2)

        for f in [real_f, complex_f, int_f, uint_f]:
            self.assertTrue(f.is_on_gpu)

    def test_device_field_set_zero(self):
        """Test that set_zero works on device fields."""
        field = self.fc.register_real_field("zero_test", 3)
        # Should not raise
        field.set_zero()


@unittest.skipUnless(GPU_AVAILABLE, "GPU backend not available")
class DeviceFieldFactoryTests(unittest.TestCase):
    """Test device field creation via factory functions."""

    def setUp(self):
        self.nb_grid_pts = (8, 8)
        # Create a device collection
        self.fc = muGrid.GlobalFieldCollection(
            self.nb_grid_pts,
            device=muGrid.Device.cuda()
        )

    def test_real_field_on_device_collection(self):
        """Test creating real field on device collection."""
        field = self.fc.real_field("device_real", (3,))
        self.assertTrue(field.is_on_gpu)

    def test_all_field_types_device(self):
        """Test all field factory functions on device collection."""
        real_f = self.fc.real_field("real_dev")
        int_f = self.fc.int_field("int_dev")
        uint_f = self.fc.uint_field("uint_dev")
        complex_f = self.fc.complex_field("complex_dev")

        for f in [real_f, int_f, uint_f, complex_f]:
            self.assertTrue(f.is_on_gpu)


@unittest.skipUnless(GPU_AVAILABLE and HAS_CUPY, "GPU backend or CuPy not available")
class CuPyIntegrationTests(unittest.TestCase):
    """Test CuPy integration for GPU fields."""

    def setUp(self):
        self.nb_grid_pts = (8, 8)
        # Create a device collection
        self.fc = muGrid.GlobalFieldCollection(
            self.nb_grid_pts,
            device=muGrid.Device.cuda()
        )

    def test_device_field_returns_cupy_array(self):
        """Test that device field accessors return CuPy arrays."""
        field = self.fc.real_field("cupy_test", (3,))

        # Access should return CuPy array
        arr = field.s
        self.assertIsInstance(arr, cp.ndarray)

    def test_cupy_array_shape(self):
        """Test that CuPy array has correct shape."""
        field = self.fc.real_field("shape_test", (2, 3))

        arr = field.s
        # Shape should be (2, 3, 1) + nb_grid_pts for SubPt layout
        expected_shape = (2, 3, 1) + self.nb_grid_pts
        self.assertEqual(arr.shape, expected_shape)

    def test_cupy_array_writability(self):
        """Test that CuPy arrays from device fields are writable."""
        field = self.fc.real_field("write_test", (2,))

        arr = field.s
        # Fill with values
        arr[:] = 42.0

        # Read back
        cp.testing.assert_array_equal(arr, 42.0)

    def test_cupy_numpy_round_trip(self):
        """Test data transfer between CuPy (device) and NumPy (host)."""
        # Create host collection
        fc_host = muGrid.GlobalFieldCollection(self.nb_grid_pts)

        # Create both host and device fields
        host_field = fc_host.real_field("host", (2,))
        device_field = self.fc.real_field("device", (2,))

        # Initialize host field with test data
        test_data = np.random.rand(*host_field.s.shape).astype(np.float64)
        host_field.s[...] = test_data

        # Copy host to device via CuPy
        device_arr = device_field.s
        device_arr[...] = cp.asarray(host_field.s)

        # Read back from device to host
        result = cp.asnumpy(device_arr)

        # Should match
        np.testing.assert_allclose(result, test_data)

    def test_all_accessors_return_cupy(self):
        """Test that all field accessors (s, sg, p, pg) return CuPy arrays."""
        field = self.fc.real_field("accessor_test", (2,))

        # All accessors should return CuPy arrays
        self.assertIsInstance(field.s, cp.ndarray)
        self.assertIsInstance(field.sg, cp.ndarray)
        self.assertIsInstance(field.p, cp.ndarray)
        self.assertIsInstance(field.pg, cp.ndarray)


@unittest.skipUnless(GPU_AVAILABLE and HAS_CUPY, "GPU backend or CuPy not available")
class MixedHostDeviceTests(unittest.TestCase):
    """Test scenarios with both host and device collections."""

    def setUp(self):
        self.nb_grid_pts = (8, 8)
        # Create host collection
        self.fc_host = muGrid.GlobalFieldCollection(self.nb_grid_pts)
        # Create device collection
        self.fc_device = muGrid.GlobalFieldCollection(
            self.nb_grid_pts,
            device=muGrid.Device.cuda()
        )

    def test_host_and_device_collections(self):
        """Test that host and device collections can coexist."""
        host_field = self.fc_host.register_real_field("host", 2)
        device_field = self.fc_device.register_real_field("device", 2)

        self.assertFalse(host_field.is_on_gpu)
        self.assertTrue(device_field.is_on_gpu)

        # Both should be accessible via DLPack
        np.from_dlpack(host_field)
        cp.from_dlpack(device_field)

    def test_different_array_types(self):
        """Test that host returns numpy, device returns cupy."""
        host_field = self.fc_host.real_field("host")
        device_field = self.fc_device.real_field("device")

        host_arr = host_field.s
        device_arr = device_field.s

        self.assertIsInstance(host_arr, np.ndarray)
        self.assertIsInstance(device_arr, cp.ndarray)


if __name__ == "__main__":
    unittest.main()



# Solver for the Poisson equation

import argparse
import json

try:
    import matplotlib.pyplot as plt
except ModuleNotFoundError:
    plt = None
import numpy as np

import muGrid
from muGrid import parprint
from muGrid.Solvers import conjugate_gradients

try:
    from mpi4py import MPI

    comm = muGrid.Communicator(MPI.COMM_WORLD)
except ImportError:
    comm = muGrid.Communicator()

from NuMPI.Testing.Subdivision import suggest_subdivisions

parser = argparse.ArgumentParser(
    prog="Poisson", description="Solve the Poisson equation"
)

parser.add_argument(
    "-n",
    "--nb-grid-pts",
    default=[32, 32],
    type=lambda s: [int(x) for x in s.split(",")],
    help="Grid points as nx,ny or nx,ny,nz (default: 32,32)",
)

_devices = {
    "cpu": muGrid.Device.cpu(),
    "gpu": muGrid.Device.gpu(),  # Auto-detect CUDA or ROCm
}

parser.add_argument(
    "-d",
    "--device",
    choices=_devices,
    default="cpu",
    help="Device for computation: 'cpu' or 'gpu' (auto-detect CUDA/ROCm) "
    "(default: cpu)",
)

parser.add_argument(
    "-i",
    "--maxiter",
    type=int,
    default=1000,
    help="Maximum number of CG iterations (default: 1000)",
)

parser.add_argument(
    "-p",
    "--plot",
    action="store_true",
    help="Show plot of RHS and solution (default: off)",
)

parser.add_argument(
    "-k",
    "--kernel",
    choices=["generic", "hardcoded"],
    default="hardcoded",
    help="Kernel implementation: 'generic' (sparse convolution) or "
    "'hardcoded' (optimized Laplace operator) (default: hardcoded)",
)

parser.add_argument(
    "-q",
    "--quiet",
    action="store_true",
    help="Suppress per-iteration output (default: off)",
)

parser.add_argument(
    "--json",
    action="store_true",
    help="Output results in JSON format (implies --quiet)",
)

args = parser.parse_args()

# JSON implies quiet mode
if args.json:
    args.quiet = True

if args.device == "cpu":
    import numpy as arr
else:
    import cupy as arr

device = _devices[args.device]

dim = len(args.nb_grid_pts)
if dim not in (2, 3):
    raise ValueError("Only 2D and 3D grids are supported")

s = suggest_subdivisions(dim, comm.size)

# Set up ghost layers for the stencil (1 layer in each direction)
left_ghosts = (1,) * dim
right_ghosts = (1,) * dim

decomposition = muGrid.CartesianDecomposition(
    comm, args.nb_grid_pts, s, left_ghosts, right_ghosts, device=device
)
grid_spacing = 1 / np.array(args.nb_grid_pts)  # Grid spacing

# FD-stencil for the Laplacian
if dim == 2:
    nb_stencil_pts = 5
else:
    nb_stencil_pts = 7

# Scaling factor for the Laplacian (we need the minus sign because
# the Laplace operator is negative definite, but the CG solver
# assumes a positive-definite operator)
laplace_scale = -1.0 / np.mean(grid_spacing) ** 2

# Create the Laplace operator based on the selected implementation
if args.kernel == "generic":
    # Generic sparse convolution operator
    if dim == 2:
        # 5-point stencil for 2D
        stencil = laplace_scale * np.array([[0, 1, 0], [1, -4, 1], [0, 1, 0]])
        stencil_offset = [-1, -1]
    else:
        # 7-point stencil for 3D
        stencil = np.zeros((3, 3, 3))
        stencil[1, 1, 0] = 1  # z-1
        stencil[1, 0, 1] = 1  # y-1
        stencil[0, 1, 1] = 1  # x-1
        stencil[1, 1, 1] = -6  # center
        stencil[2, 1, 1] = 1  # x+1
        stencil[1, 2, 1] = 1  # y+1
        stencil[1, 1, 2] = 1  # z+1
        stencil *= laplace_scale
        stencil_offset = [-1, -1, -1]
    laplace = muGrid.GenericLinearOperator(stencil_offset, stencil)
    stencil_name = "Generic sparse convolution"
else:
    # Hard-coded optimized Laplace operator (for benchmarking)
    # Pass scale factor to fold in grid spacing and positive-definiteness
    laplace = muGrid.LaplaceOperator(dim, laplace_scale)
    stencil_name = "Hard-coded Laplace operator"

coords = decomposition.coords  # Domain-local coords for each pixel

# Create fields using the decomposition's method API
rhs = decomposition.real_field("rhs")
solution = decomposition.real_field("solution")

# Set up RHS with a smooth function
if dim == 2:
    x, y = coords
    rhs.p[...] = arr.asarray((1 + np.cos(2 * np.pi * x) * np.cos(2 * np.pi * y)) ** 10)
else:
    x, y, z = coords
    rhs.p[...] = arr.asarray(
        (1 + np.cos(2 * np.pi * x) * np.cos(2 * np.pi * y) * np.cos(2 * np.pi * z))
        ** 10
    )
rhs.p[...] -= arr.mean(rhs.p)

# Performance counters
nb_grid_pts_total = np.prod(args.nb_grid_pts)

# Create global timer for hierarchical timing
timer = muGrid.Timer()


def callback(iteration, state):
    """
    Callback function to print the iteration and squared residual norm.
    """
    if not args.quiet:
        parprint(f"{iteration:5} {state['rr']:.5}", comm=comm)


def hessp(x, Ax):
    """
    Hessian-vector product function for CG solver.

    This function computes the product of the Hessian matrix with a vector.
    """
    with timer("communicate_ghosts"):
        decomposition.communicate_ghosts(x)
    with timer("apply"):
        laplace.apply(x, Ax)


converged = False
with timer("conjugate_gradients"):
    try:
        conjugate_gradients(
            comm,
            decomposition,
            rhs,
            solution,
            hessp=hessp,
            tol=1e-6,
            callback=callback,
            maxiter=args.maxiter,
            timer=timer,
        )
        converged = True
        if not args.quiet:
            parprint("CG converged.", comm=comm)
    except RuntimeError:
        if not args.quiet:
            parprint("CG did not converge.", comm=comm)

elapsed_time = timer.get_time("conjugate_gradients")

# Performance metrics calculations
# Get number of hessp calls from timer (= number of CG iterations)
nb_iterations = timer.get_calls("conjugate_gradients/iteration/hessp")

# Lattice updates per second (LUPS)
# One "lattice update" = one grid point processed in one CG iteration
total_lattice_updates = nb_grid_pts_total * nb_iterations
lups = total_lattice_updates / elapsed_time if elapsed_time > 0 else 0

# Memory and FLOPS estimates per CG iteration:
# Each value is 8 bytes (double precision)
#
# Per CG iteration (excluding ghost communication):
#   hessp (apply):         read nb_stencil_pts, write 1, FLOPs 2*nb_stencil_pts
#   dot_pAp:               read 2 (p, Ap),       FLOPs 2 (mul + add)
#   update_x (axpy):       read 2, write 1,      FLOPs 2
#   update_r (axpy_norm_sq): read 2, write 1,    FLOPs 4 (fused axpy + norm)
#   update_p (axpby):      read 2, write 1,      FLOPs 2 (mul + mul + add)
#
# The fused axpy_norm_sq saves 1 read compared to separate axpy + norm_sq,
# because the norm is computed during the write pass without re-reading y.
#
# Total reads:  nb_stencil_pts + 2 + 2 + 2 + 2 = nb_stencil_pts + 8
# Total writes: 1 + 0 + 1 + 1 + 1 = 4
# Total FLOPs:  2*nb_stencil_pts + 2 + 2 + 4 + 2 = 2*nb_stencil_pts + 10

reads_per_iteration = nb_stencil_pts + 8  # values read per grid point
writes_per_iteration = 4  # values written per grid point
flops_per_iteration = 2 * nb_stencil_pts + 10  # FLOPs per grid point

bytes_per_iteration = (
    nb_grid_pts_total * (reads_per_iteration + writes_per_iteration) * 8
)
total_bytes = nb_iterations * bytes_per_iteration
memory_throughput = total_bytes / elapsed_time if elapsed_time > 0 else 0

flops_per_cg_iteration = nb_grid_pts_total * flops_per_iteration
total_flops = nb_iterations * flops_per_cg_iteration
flops_rate = total_flops / elapsed_time if elapsed_time > 0 else 0

# Arithmetic intensity (FLOPs per byte)
arithmetic_intensity = flops_per_iteration / (
    (reads_per_iteration + writes_per_iteration) * 8
)

# Breakdown: hessp (apply) only
bytes_per_hessp = nb_grid_pts_total * (nb_stencil_pts + 1) * 8
flops_per_hessp = nb_grid_pts_total * (2 * nb_stencil_pts)
apply_time = timer.get_time("conjugate_gradients/iteration/hessp/apply")
apply_lups = total_lattice_updates / apply_time if apply_time > 0 else 0
apply_throughput = (
    (nb_iterations * bytes_per_hessp) / apply_time if apply_time > 0 else 0
)
apply_flops_rate = (
    (nb_iterations * flops_per_hessp) / apply_time if apply_time > 0 else 0
)

if args.json:
    # JSON output (convert numpy types to Python types for JSON serialization)
    # Timer's to_dict() includes PAPI data when available
    results = {
        "config": {
            "nb_grid_pts": [int(x) for x in args.nb_grid_pts],
            "nb_grid_pts_total": int(nb_grid_pts_total),
            "dimensions": int(dim),
            "kernel": args.kernel,
            "stencil_name": stencil_name,
            "nb_stencil_pts": int(nb_stencil_pts),
            "device": device.device_string,
            "maxiter": int(args.maxiter),
        },
        "results": {
            "converged": converged,
            "iterations": int(nb_iterations),
            "total_time_seconds": float(elapsed_time),
            "total_lattice_updates": int(total_lattice_updates),
            "MLUPS": float(lups / 1e6),
            "GLUPS": float(lups / 1e9),
            "reads_per_grid_point": int(reads_per_iteration),
            "writes_per_grid_point": int(writes_per_iteration),
            "bytes_per_iteration": int(bytes_per_iteration),
            "total_bytes": int(total_bytes),
            "memory_throughput_GBps": float(memory_throughput / 1e9),
            "flops_per_grid_point": int(flops_per_iteration),
            "flops_per_iteration": int(flops_per_cg_iteration),
            "total_flops": int(total_flops),
            "flops_rate_GFLOPs": float(flops_rate / 1e9),
            "arithmetic_intensity": float(arithmetic_intensity),
            "apply_time_seconds": float(apply_time),
            "apply_MLUPS": float(apply_lups / 1e6),
            "apply_throughput_GBps": float(apply_throughput / 1e9),
            "apply_flops_rate_GFLOPs": float(apply_flops_rate / 1e9),
        },
        "timing": timer.to_dict(),
    }
    parprint(json.dumps(results, indent=2), comm=comm)
else:
    # Text output
    parprint(f"\n{'='*60}", comm=comm)
    parprint("Performance Summary", comm=comm)
    parprint(f"{'='*60}", comm=comm)
    parprint(
        f"Grid size: {' x '.join(map(str, args.nb_grid_pts))} = "
        f"{nb_grid_pts_total:,} points",
        comm=comm,
    )
    parprint(f"Dimensions: {dim}D", comm=comm)
    parprint(f"Device: {device.device_string}", comm=comm)
    parprint(f"Stencil implementation: {stencil_name}", comm=comm)
    parprint(f"Stencil points: {nb_stencil_pts}", comm=comm)
    parprint(f"CG iterations: {nb_iterations}", comm=comm)
    parprint(f"Total time: {elapsed_time:.4f} seconds", comm=comm)

    parprint("\nLattice updates per second:", comm=comm)
    parprint(f"  Total lattice updates: {total_lattice_updates:,}", comm=comm)
    parprint(f"  LUPS: {lups / 1e6:.2f} MLUPS ({lups / 1e9:.4f} GLUPS)", comm=comm)

    parprint("\nMemory traffic per CG iteration (estimated):", comm=comm)
    parprint(
        f"  Per grid point: {reads_per_iteration} reads + "
        f"{writes_per_iteration} writes "
        f"= {(reads_per_iteration + writes_per_iteration) * 8} bytes",
        comm=comm,
    )
    parprint(f"    hessp:    {nb_stencil_pts} reads, 1 write", comm=comm)
    parprint("    dot_pAp:  2 reads", comm=comm)
    parprint("    update_x: 2 reads, 1 write", comm=comm)
    parprint("    update_r: 2 reads, 1 write (fused axpy_norm_sq)", comm=comm)
    parprint("    update_p: 2 reads, 1 write", comm=comm)
    parprint(f"  Per iteration: {bytes_per_iteration / 1e6:.2f} MB", comm=comm)
    parprint(f"  Total: {total_bytes / 1e9:.2f} GB", comm=comm)
    parprint(f"  Throughput: {memory_throughput / 1e9:.2f} GB/s", comm=comm)

    parprint("\nFLOPs per CG iteration (estimated):", comm=comm)
    parprint(f"  Per grid point: {flops_per_iteration} FLOPs", comm=comm)
    parprint(f"    hessp:    {2 * nb_stencil_pts} FLOPs", comm=comm)
    parprint("    dot_pAp:  2 FLOPs", comm=comm)
    parprint("    update_x: 2 FLOPs", comm=comm)
    parprint("    update_r: 4 FLOPs (fused axpy_norm_sq)", comm=comm)
    parprint("    update_p: 2 FLOPs", comm=comm)
    parprint(f"  Per iteration: {flops_per_cg_iteration / 1e6:.2f} MFLOP", comm=comm)
    parprint(f"  Total: {total_flops / 1e9:.2f} GFLOP", comm=comm)
    parprint(f"  FLOP rate: {flops_rate / 1e9:.2f} GFLOP/s", comm=comm)

    parprint(f"\nArithmetic intensity: {arithmetic_intensity:.3f} FLOP/byte", comm=comm)
    parprint(f"{'='*60}", comm=comm)

    # Print hierarchical timing breakdown (includes PAPI data when enabled)
    timer.print_summary()

if args.plot:
    if dim == 3:
        parprint("Warning: Plotting not supported for 3D grids", comm=comm)
    elif plt is None:
        parprint("Warning: matplotlib not available, cannot show plot", comm=comm)
    else:
        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(10, 5))
        ax1.imshow(rhs.p)
        ax1.set_title("RHS")
        ax2.imshow(solution.p)
        ax2.set_title("Solution")
        plt.tight_layout()
        plt.show()



import numpy as np
import muGrid

engine = muGrid.FFTEngine([16, 20])

real_field = engine.real_space_field("real")
fourier_field = engine.fourier_space_field("fourier")

# Initialize real-space field
real_field.s[:] = np.random.randn(1, 1, 16, 20)
original = real_field.s.copy()

# Forward FFT: real -> Fourier
engine.fft(real_field, fourier_field)

# Inverse FFT: Fourier -> real
engine.ifft(fourier_field, real_field)

# Apply normalization for roundtrip
real_field.s[:] *= engine.normalisation

# Verify roundtrip
np.testing.assert_allclose(real_field.s, original, atol=1e-14)



#!/usr/bin/env python3
# -*- coding:utf-8 -*-
"""
@file     python_fft_tests.py

@author  Lars Pastewka <lars.pastewka@imtek.uni-freiburg.de>

@date    18 Dec 2024

@brief   test FFT engine Python bindings

Copyright © 2024 Lars Pastewka

µGrid is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public License as
published by the Free Software Foundation, either version 3, or (at
your option) any later version.

µGrid is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public License
along with µGrid; see the file COPYING. If not, write to the
Free Software Foundation, Inc., 59 Temple Place - Suite 330,
Boston, MA 02111-1307, USA.

Additional permission under GNU GPL version 3 section 7

If you modify this Program, or any covered work, by linking or combining it
with proprietary FFT implementations or numerical libraries, containing parts
covered by the terms of those libraries' licenses, the licensors of this
Program grant you additional permission to convey the resulting work.
"""

import unittest

import numpy as np
import pytest
from conftest import get_array_module, get_test_devices, skip_if_gpu_unavailable
from numpy.testing import assert_allclose, assert_array_equal

import muGrid


class FFTUtilsTest(unittest.TestCase):
    """Test FFT utility functions."""

    def test_get_hermitian_grid_pts_2d(self):
        """Test hermitian grid dimensions for 2D."""
        fourier = muGrid.get_hermitian_grid_pts([8, 10])
        self.assertEqual(fourier[0], 5)  # 8/2 + 1
        self.assertEqual(fourier[1], 10)

    def test_get_hermitian_grid_pts_3d(self):
        """Test hermitian grid dimensions for 3D."""
        fourier = muGrid.get_hermitian_grid_pts([8, 10, 12])
        self.assertEqual(fourier[0], 5)  # 8/2 + 1
        self.assertEqual(fourier[1], 10)
        self.assertEqual(fourier[2], 12)

    def test_fft_normalization(self):
        """Test FFT normalization factor."""
        norm_2d = muGrid.fft_normalization([8, 10])
        self.assertAlmostEqual(norm_2d, 1.0 / 80, places=15)

        norm_3d = muGrid.fft_normalization([8, 10, 12])
        self.assertAlmostEqual(norm_3d, 1.0 / 960, places=15)


class FFTEngineCreationTest(unittest.TestCase):
    """Test FFT engine creation and properties."""

    def test_create_2d_engine(self):
        """Test creating a 2D FFT engine."""
        engine = muGrid.FFTEngine([16, 20])

        # Check real space dimensions
        self.assertEqual(engine.nb_domain_grid_pts[0], 16)
        self.assertEqual(engine.nb_domain_grid_pts[1], 20)

        # Check Fourier space dimensions (r2c transform)
        self.assertEqual(engine.nb_fourier_grid_pts[0], 9)  # 16/2 + 1
        self.assertEqual(engine.nb_fourier_grid_pts[1], 20)

    def test_create_3d_engine(self):
        """Test creating a 3D FFT engine."""
        engine = muGrid.FFTEngine([8, 10, 12])

        self.assertEqual(engine.nb_domain_grid_pts[0], 8)
        self.assertEqual(engine.nb_domain_grid_pts[1], 10)
        self.assertEqual(engine.nb_domain_grid_pts[2], 12)

        self.assertEqual(engine.nb_fourier_grid_pts[0], 5)  # 8/2 + 1
        self.assertEqual(engine.nb_fourier_grid_pts[1], 10)
        self.assertEqual(engine.nb_fourier_grid_pts[2], 12)

    def test_normalization(self):
        """Test normalization factor."""
        engine_2d = muGrid.FFTEngine([8, 10])
        self.assertAlmostEqual(engine_2d.normalisation, 1.0 / 80, places=15)

        engine_3d = muGrid.FFTEngine([8, 10, 12])
        self.assertAlmostEqual(engine_3d.normalisation, 1.0 / 960, places=15)

    def test_backend_name(self):
        """Test backend name is reported."""
        engine = muGrid.FFTEngine([8, 10])
        self.assertEqual(engine.backend_name, "PocketFFT")


class FFTFieldTest(unittest.TestCase):
    """Test FFT field creation and access."""

    def setUp(self):
        self.engine_2d = muGrid.FFTEngine([16, 20])
        self.engine_3d = muGrid.FFTEngine([8, 10, 12])

    def test_create_real_field(self):
        """Test creating real-space field."""
        field = self.engine_2d.real_space_field("test_real")
        self.assertEqual(field.name, "test_real")
        self.assertEqual(field.s.shape, (1, 1, 16, 20))

    def test_create_fourier_field(self):
        """Test creating Fourier-space field."""
        field = self.engine_2d.fourier_space_field("test_fourier")
        self.assertEqual(field.name, "test_fourier")
        self.assertEqual(field.s.shape, (1, 1, 9, 20))  # Half-complex

    def test_field_data_types(self):
        """Test field data types."""
        real_field = self.engine_2d.real_space_field("real")
        fourier_field = self.engine_2d.fourier_space_field("fourier")

        self.assertEqual(real_field.s.dtype, np.float64)
        self.assertEqual(fourier_field.s.dtype, np.complex128)

    def test_field_pixel_layout(self):
        """Test pixel layout access."""
        real_field = self.engine_2d.real_space_field("real_p")

        # SubPt layout: (components, sub_pts, nx, ny)
        self.assertEqual(real_field.s.shape, (1, 1, 16, 20))

        # Pixel layout: (components * sub_pts, nx, ny)
        self.assertEqual(real_field.p.shape, (1, 16, 20))

    def test_field_3d_shapes(self):
        """Test 3D field shapes."""
        real_field = self.engine_3d.real_space_field("real_3d")
        fourier_field = self.engine_3d.fourier_space_field("fourier_3d")

        self.assertEqual(real_field.s.shape, (1, 1, 8, 10, 12))
        self.assertEqual(fourier_field.s.shape, (1, 1, 5, 10, 12))

    def test_field_write_access(self):
        """Test that field data can be written."""
        field = self.engine_2d.real_space_field("writable")
        field.s[:] = 1.0
        self.assertTrue(np.all(field.s == 1.0))

        field.p[:] = 2.0
        self.assertTrue(np.all(field.p == 2.0))

    def test_real_space_field_returns_existing(self):
        """Test that real_space_field returns existing field if present."""
        field1 = self.engine_2d.real_space_field("shared_real")
        field1.s[:] = 42.0

        # Calling again with same name should return the same field
        field2 = self.engine_2d.real_space_field("shared_real")

        # Both should reference the same data
        self.assertTrue(np.all(field2.s == 42.0))

        # Modifying field2 should affect field1
        field2.s[:] = 99.0
        self.assertTrue(np.all(field1.s == 99.0))

    def test_fourier_space_field_returns_existing(self):
        """Test that fourier_space_field returns existing field if present."""
        field1 = self.engine_2d.fourier_space_field("shared_fourier")
        field1.s[:] = 42.0 + 0j

        # Calling again with same name should return the same field
        field2 = self.engine_2d.fourier_space_field("shared_fourier")

        # Both should reference the same data
        self.assertTrue(np.all(field2.s == 42.0 + 0j))

        # Modifying field2 should affect field1
        field2.s[:] = 99.0 + 1j
        self.assertTrue(np.all(field1.s == 99.0 + 1j))

    def test_register_real_space_field_throws_on_duplicate(self):
        """Test that register_real_space_field throws if field exists."""
        self.engine_2d.register_real_space_field("unique_real")

        # Registering again should raise
        with self.assertRaises(RuntimeError):
            self.engine_2d.register_real_space_field("unique_real")

    def test_register_fourier_space_field_throws_on_duplicate(self):
        """Test that register_fourier_space_field throws if field exists."""
        self.engine_2d.register_fourier_space_field("unique_fourier")

        # Registering again should raise
        with self.assertRaises(RuntimeError):
            self.engine_2d.register_fourier_space_field("unique_fourier")


class FFTRoundtripTest(unittest.TestCase):
    """Test FFT forward/inverse roundtrip."""

    def test_2d_roundtrip_sine(self):
        """Test 2D FFT roundtrip with sine wave."""
        engine = muGrid.FFTEngine([16, 20])
        real_field = engine.real_space_field("real")
        fourier_field = engine.fourier_space_field("fourier")

        # Initialize with sine wave
        x = np.arange(16 * 20).reshape(1, 1, 16, 20)
        real_field.s[:] = np.sin(2 * np.pi * x / (16 * 20))
        original = real_field.s.copy()

        # Forward and inverse FFT
        engine.fft(real_field, fourier_field)
        engine.ifft(fourier_field, real_field)

        # Normalize
        real_field.s[:] *= engine.normalisation

        # Check roundtrip
        assert_allclose(real_field.s, original, atol=1e-14)

    def test_2d_roundtrip_random(self):
        """Test 2D FFT roundtrip with random data."""
        engine = muGrid.FFTEngine([16, 20])
        real_field = engine.real_space_field("real")
        fourier_field = engine.fourier_space_field("fourier")

        # Initialize with random data
        np.random.seed(42)
        real_field.s[:] = np.random.randn(1, 1, 16, 20)
        original = real_field.s.copy()

        # Forward and inverse FFT
        engine.fft(real_field, fourier_field)
        engine.ifft(fourier_field, real_field)
        real_field.s[:] *= engine.normalisation

        # Check roundtrip
        assert_allclose(real_field.s, original, atol=1e-14)

    def test_3d_roundtrip(self):
        """Test 3D FFT roundtrip."""
        engine = muGrid.FFTEngine([8, 10, 12])
        real_field = engine.real_space_field("real")
        fourier_field = engine.fourier_space_field("fourier")

        # Initialize
        np.random.seed(42)
        real_field.s[:] = np.random.randn(1, 1, 8, 10, 12)
        original = real_field.s.copy()

        # Forward and inverse FFT
        engine.fft(real_field, fourier_field)
        engine.ifft(fourier_field, real_field)
        real_field.s[:] *= engine.normalisation

        # Check roundtrip
        assert_allclose(real_field.s, original, atol=1e-14)

    def test_fft_dc_component(self):
        """Test that constant field has only DC component."""
        engine = muGrid.FFTEngine([8, 10])
        real_field = engine.real_space_field("real")
        fourier_field = engine.fourier_space_field("fourier")

        # Constant field
        real_field.s[:] = 5.0

        engine.fft(real_field, fourier_field)

        # DC component should be sum = 5 * 80 = 400
        self.assertAlmostEqual(fourier_field.s[0, 0, 0, 0].real, 400.0, places=10)
        self.assertAlmostEqual(fourier_field.s[0, 0, 0, 0].imag, 0.0, places=10)

        # All other components should be zero
        fourier_copy = fourier_field.s.copy()
        fourier_copy[0, 0, 0, 0] = 0
        assert_allclose(fourier_copy, 0, atol=1e-12)


def _get_device_string(device):
    """Convert device parameter to device string for FFTEngine."""
    if device == "cpu":
        return None
    else:
        return "cuda:0"


@pytest.mark.parametrize("device", get_test_devices())
class TestFFTMultiComponent:
    """Test FFT with multi-component (vector/tensor) fields."""

    def test_2d_vector_field_roundtrip(self, device):
        """Test 2D FFT roundtrip with 2-component vector field."""
        skip_if_gpu_unavailable(device)
        xp = get_array_module(device)

        engine = muGrid.FFTEngine([16, 20], device=_get_device_string(device))
        real_field = engine.real_space_field("vector", nb_components=2)
        fourier_field = engine.fourier_space_field("vector_k", nb_components=2)

        # Initialize with different patterns for each component
        np.random.seed(42)
        real_field.p[0, ...] = xp.asarray(np.random.randn(16, 20))
        real_field.p[1, ...] = xp.asarray(np.random.randn(16, 20))
        original = real_field.p.copy()

        # Forward and inverse FFT
        engine.fft(real_field, fourier_field)
        engine.ifft(fourier_field, real_field)

        # Normalize
        real_field.p[:] *= engine.normalisation

        # Check roundtrip for both components
        result = real_field.p
        if device == "gpu":
            result = result.get()
            original = original.get()
        assert_allclose(result, original, atol=1e-14)

    def test_2d_tensor_field_roundtrip(self, device):
        """Test 2D FFT roundtrip with 4-component (2x2 tensor) field."""
        skip_if_gpu_unavailable(device)
        xp = get_array_module(device)

        engine = muGrid.FFTEngine([16, 20], device=_get_device_string(device))
        real_field = engine.real_space_field("tensor", nb_components=4)
        fourier_field = engine.fourier_space_field("tensor_k", nb_components=4)

        # Initialize with different patterns for each component
        np.random.seed(123)
        for c in range(4):
            real_field.p[c, ...] = xp.asarray(np.random.randn(16, 20))
        original = real_field.p.copy()

        # Forward and inverse FFT
        engine.fft(real_field, fourier_field)
        engine.ifft(fourier_field, real_field)

        # Normalize
        real_field.p[:] *= engine.normalisation

        # Check roundtrip for all components
        result = real_field.p
        if device == "gpu":
            result = result.get()
            original = original.get()
        assert_allclose(result, original, atol=1e-14)

    def test_3d_vector_field_roundtrip(self, device):
        """Test 3D FFT roundtrip with 3-component vector field."""
        skip_if_gpu_unavailable(device)
        xp = get_array_module(device)

        engine = muGrid.FFTEngine([8, 10, 12], device=_get_device_string(device))
        real_field = engine.real_space_field("vector3d", nb_components=3)
        fourier_field = engine.fourier_space_field("vector3d_k", nb_components=3)

        # Initialize with different patterns for each component
        np.random.seed(456)
        for c in range(3):
            real_field.p[c, ...] = xp.asarray(np.random.randn(8, 10, 12))
        original = real_field.p.copy()

        # Forward and inverse FFT
        engine.fft(real_field, fourier_field)
        engine.ifft(fourier_field, real_field)

        # Normalize
        real_field.p[:] *= engine.normalisation

        # Check roundtrip for all components
        result = real_field.p
        if device == "gpu":
            result = result.get()
            original = original.get()
        assert_allclose(result, original, atol=1e-14)

    def test_vector_field_dc_component(self, device):
        """Test that constant vector field has only DC components."""
        skip_if_gpu_unavailable(device)

        engine = muGrid.FFTEngine([8, 10], device=_get_device_string(device))
        real_field = engine.real_space_field("vec", nb_components=2)
        fourier_field = engine.fourier_space_field("vec_k", nb_components=2)

        # Constant field: component 0 = 3.0, component 1 = 5.0
        real_field.p[0, ...] = 3.0
        real_field.p[1, ...] = 5.0

        engine.fft(real_field, fourier_field)

        # Get result on CPU for comparison
        fourier_p = fourier_field.p
        if device == "gpu":
            fourier_p = fourier_p.get()

        # DC components should be sum = value * 80
        # Component 0: 3 * 80 = 240
        # Component 1: 5 * 80 = 400
        assert abs(fourier_p[0, 0, 0].real - 240.0) < 1e-10
        assert abs(fourier_p[0, 0, 0].imag) < 1e-10
        assert abs(fourier_p[1, 0, 0].real - 400.0) < 1e-10
        assert abs(fourier_p[1, 0, 0].imag) < 1e-10

        # All other Fourier coefficients should be zero
        fourier_copy = fourier_p.copy()
        fourier_copy[:, 0, 0] = 0
        assert_allclose(fourier_copy, 0, atol=1e-12)

    def test_components_independent(self, device):
        """Test that FFT of each component is independent."""
        skip_if_gpu_unavailable(device)
        xp = get_array_module(device)

        engine = muGrid.FFTEngine([16, 20], device=_get_device_string(device))

        # Multi-component field
        real_multi = engine.real_space_field("multi", nb_components=2)
        fourier_multi = engine.fourier_space_field("multi_k", nb_components=2)

        # Single component fields for reference
        real_single0 = engine.real_space_field("single0", nb_components=1)
        fourier_single0 = engine.fourier_space_field("single0_k", nb_components=1)
        real_single1 = engine.real_space_field("single1", nb_components=1)
        fourier_single1 = engine.fourier_space_field("single1_k", nb_components=1)

        # Initialize with same data
        np.random.seed(789)
        data0 = np.random.randn(16, 20)
        data1 = np.random.randn(16, 20)

        real_multi.p[0, ...] = xp.asarray(data0)
        real_multi.p[1, ...] = xp.asarray(data1)
        real_single0.p[0, ...] = xp.asarray(data0)
        real_single1.p[0, ...] = xp.asarray(data1)

        # FFT all fields
        engine.fft(real_multi, fourier_multi)
        engine.fft(real_single0, fourier_single0)
        engine.fft(real_single1, fourier_single1)

        # Get results on CPU for comparison
        fourier_multi_p = fourier_multi.p
        fourier_single0_p = fourier_single0.p
        fourier_single1_p = fourier_single1.p
        if device == "gpu":
            fourier_multi_p = fourier_multi_p.get()
            fourier_single0_p = fourier_single0_p.get()
            fourier_single1_p = fourier_single1_p.get()

        # Multi-component FFT should match single-component FFTs
        assert_allclose(fourier_multi_p[0, ...], fourier_single0_p[0, ...], atol=1e-14)
        assert_allclose(fourier_multi_p[1, ...], fourier_single1_p[0, ...], atol=1e-14)


class FFTFrequencyTest(unittest.TestCase):
    """Test FFT frequency and coordinate properties."""

    def test_fftfreq_2d(self):
        """Test 2D fftfreq property matches numpy."""
        nb_grid_pts = [7, 4]
        nx, ny = nb_grid_pts
        engine = muGrid.FFTEngine(nb_grid_pts)

        # Build reference from numpy
        freq_ref = np.array(
            np.meshgrid(*(np.fft.fftfreq(n) for n in nb_grid_pts), indexing="ij")
        )
        # Slice for half-complex (r2c) transform
        freq_ref = freq_ref[:, : nx // 2 + 1, :]

        assert_allclose(engine.fftfreq, freq_ref)

    def test_ifftfreq_2d(self):
        """Test 2D ifftfreq property (integer indices)."""
        nb_grid_pts = [7, 4]
        nx, ny = nb_grid_pts
        engine = muGrid.FFTEngine(nb_grid_pts)

        # Build reference from numpy
        freq_ref = np.array(
            np.meshgrid(
                *(np.fft.fftfreq(n, 1 / n) for n in nb_grid_pts),
                indexing="ij",
            )
        )
        freq_ref = freq_ref[:, : nx // 2 + 1, :]

        assert_allclose(engine.ifftfreq, freq_ref)

    def test_fftfreq_3d(self):
        """Test 3D fftfreq property matches numpy."""
        nb_grid_pts = [6, 4, 5]
        nx, ny, nz = nb_grid_pts
        engine = muGrid.FFTEngine(nb_grid_pts)

        # Build reference from numpy
        freq_ref = np.array(
            np.meshgrid(*(np.fft.fftfreq(n) for n in nb_grid_pts), indexing="ij")
        )
        # Slice for half-complex (r2c) transform
        freq_ref = freq_ref[:, : nx // 2 + 1, :, :]

        assert_allclose(engine.fftfreq, freq_ref)

    def test_coords_2d(self):
        """Test 2D coords property."""
        nb_grid_pts = [7, 4]
        nx, ny = nb_grid_pts
        engine = muGrid.FFTEngine(nb_grid_pts)

        x, y = engine.coords
        xref, yref = np.mgrid[0:nx, 0:ny]

        assert_allclose(x, xref / nx)
        assert_allclose(y, yref / ny)

    def test_coords_3d(self):
        """Test 3D coords property."""
        nb_grid_pts = [7, 4, 5]
        nx, ny, nz = nb_grid_pts
        engine = muGrid.FFTEngine(nb_grid_pts)

        assert_array_equal(engine.coords.shape, [3] + nb_grid_pts)

        x, y, z = engine.coords
        xref, yref, zref = np.mgrid[0:nx, 0:ny, 0:nz]

        assert_allclose(x, xref / nx)
        assert_allclose(y, yref / ny)
        assert_allclose(z, zref / nz)

    def test_icoords_2d(self):
        """Test 2D icoords property (integer indices)."""
        nb_grid_pts = [7, 4]
        nx, ny = nb_grid_pts
        engine = muGrid.FFTEngine(nb_grid_pts)

        x, y = engine.coords
        ix, iy = engine.icoords

        # Integer coords should equal fractional * n
        assert_allclose(ix, x * nx)
        assert_allclose(iy, y * ny)

    def test_icoords_dtype(self):
        """Test that icoords returns integer type."""
        engine = muGrid.FFTEngine([7, 4])
        ix, iy = engine.icoords
        self.assertTrue(np.issubdtype(ix.dtype, np.integer))
        self.assertTrue(np.issubdtype(iy.dtype, np.integer))

    def test_ifftfreq_dtype(self):
        """Test that ifftfreq returns integer type."""
        engine = muGrid.FFTEngine([7, 4])
        iqx, iqy = engine.ifftfreq
        self.assertTrue(np.issubdtype(iqx.dtype, np.integer))
        self.assertTrue(np.issubdtype(iqy.dtype, np.integer))

    def test_frequency_coordinate_consistency(self):
        """Test that setting Fourier modes using fftfreq produces correct result."""
        nb_grid_pts = [7, 4]
        nx, ny = nb_grid_pts
        engine = muGrid.FFTEngine(nb_grid_pts)

        x, y = engine.coords
        qx, qy = engine.fftfreq

        real_field = engine.real_space_field("real")
        fourier_field = engine.fourier_space_field("fourier")

        # Set a single x-direction mode
        fourier_field.s[:] = 0
        fourier_field.s[
            0,
            0,
            np.logical_and(
                np.abs(np.abs(qx) * nx - 1) < 1e-6, np.abs(np.abs(qy) * ny - 0) < 1e-6
            ),
        ] = 0.5

        engine.ifft(fourier_field, real_field)
        # Should produce cos(2*pi*x)
        assert_allclose(real_field.p[0], np.cos(2 * np.pi * x), atol=1e-12)

        # Set a single y-direction mode
        fourier_field.s[:] = 0
        fourier_field.s[
            0,
            0,
            np.logical_and(
                np.abs(np.abs(qx) * nx - 0) < 1e-6, np.abs(np.abs(qy) * ny - 1) < 1e-6
            ),
        ] = 0.5

        engine.ifft(fourier_field, real_field)
        # Should produce cos(2*pi*y)
        assert_allclose(real_field.p[0], np.cos(2 * np.pi * y), atol=1e-12)

    def test_properties_return_tuples(self):
        """Test that dimension properties return tuples."""
        engine = muGrid.FFTEngine([8, 10])

        # These should all be tuples, not DynGridIndex objects
        self.assertIsInstance(engine.nb_fourier_grid_pts, tuple)
        self.assertIsInstance(engine.nb_fourier_subdomain_grid_pts, tuple)
        self.assertIsInstance(engine.fourier_subdomain_locations, tuple)
        self.assertIsInstance(engine.nb_subdomain_grid_pts, tuple)
        self.assertIsInstance(engine.subdomain_locations, tuple)
        self.assertIsInstance(engine.process_grid, tuple)
        self.assertIsInstance(engine.process_coords, tuple)

    def test_spatial_dim_property(self):
        """Test spatial_dim property."""
        engine_2d = muGrid.FFTEngine([8, 10])
        self.assertEqual(engine_2d.spatial_dim, 2)

        engine_3d = muGrid.FFTEngine([8, 10, 12])
        self.assertEqual(engine_3d.spatial_dim, 3)


if __name__ == "__main__":
    unittest.main()
